from collections import defaultdict, OrderedDict
from contextlib import contextmanager
import typing
from typing import (
    Any, Optional,
    TypeVar, Union, TYPE_CHECKING
)
from collections.abc import AsyncGenerator, Generator, Iterator, Mapping, Sequence, Iterable
from typing_extensions import Protocol, Literal
from typing import runtime_checkable
import warnings
import logging
import uuid

import cloudpickle
import numpy as np
from opentelemetry import trace, context as opentelemetry_context
from sparseconverter import (
    BACKENDS, CUPY_BACKENDS, D2_BACKENDS, DENSE_BACKENDS, ND_BACKENDS, SPARSE_BACKENDS,
    CUDA_BACKENDS, CPU_BACKENDS, CUDA,
    ArrayBackend, cheapest_pair, check_shape, get_backend,
    make_like, result_type, for_backend
)

from libertem.common.tracing import attach_to_parent, TracedThreadPoolExecutor
from libertem.common.progress import ProgressManager, PartitionProgressTracker, PartitionTrackerNoOp
from libertem.io.dataset.base.tiling import DataTile
from libertem.io.dataset.base import DataSetMeta
from libertem.utils.devices import has_cupy
from libertem.warnings import UseDiscouragedWarning
from libertem.common.exceptions import UDFException
from libertem.common.buffers import (
    BufferWrapper, AuxBufferWrapper, PlaceholderBufferWrapper,
    BufferKind, BufferUse, BufferLocation, ArrayWithMask,
)
from libertem.common import Shape, Slice
from libertem.common.udf import TilingPreferences, UDFProtocol, UDFMethod
from libertem.common.math import count_nonzero, prod
from libertem.io.dataset.base import (
    TilingScheme, Negotiator, Partition, DataSet, get_coordinates
)
from libertem.io.corrections import CorrectionSet
from libertem.io.dataset.base.roi import roi_for_partition
from libertem.common.backend import get_use_cuda, get_device_class
from libertem.common.async_utils import async_generator_eager
from libertem.executor.inline import InlineJobExecutor
from libertem.common.executor import (
    Environment, JobExecutor, TaskCommHandler, NoopCommHandler, TaskProtocol,
    JobCancelledError, ResourceDef,
)
from libertem.common.exceptions import UDFRunCancelled

if TYPE_CHECKING:
    from numpy import typing as nt
    from opentelemetry.trace import SpanContext
    from libertem.common.progress import ProgressReporter

tracer = trace.get_tracer(__name__)
log = logging.getLogger(__name__)

BackendSpec = Union[ArrayBackend, Iterable[ArrayBackend]]
DeviceClass = Literal['cpu', 'cuda']
UDFKwarg = Union[Any, AuxBufferWrapper]
UDFKwargs = dict[str, UDFKwarg]
ExecutionPlan = dict[ArrayBackend, Iterable["UDF"]]


class ResultsForDataSet:
    def __init__(
        self,
        gen: Generator[tuple[tuple["UDFData", ...], TaskProtocol], None, None],
        params: "UDFParams",
        params_handle,
        executor: JobExecutor,
        udfs: list["UDF"],
    ):
        self._gen = gen
        self._params = params
        self._params_handle = params_handle
        self._executor = executor
        self._udfs = udfs

    def __iter__(self):
        return self

    def __next__(self) -> tuple[tuple["UDFData", ...], TaskProtocol]:
        return next(self._gen)

    def close(self):
        self._gen.close()

    def throw(self, exc: Exception):
        return self._gen.throw(exc)

    def update_parameters_experimental(self, patch: list[dict[str, Any]]):
        log.debug("ResultsForDataSet.update_parameters_experimental: %s", patch)
        # update parameters on workers (via the executor):
        self._params.patch(patch=patch)
        self._executor.scatter_update_patch(self._params_handle, patch)
        # update UDFs on the main node:
        for udf, params in zip(self._udfs, patch):
            udf._update_parameters_patch(kwargs_patch=params)


def _get_dtype(
    udfs: list["UDF"],
    dtype: "nt.DTypeLike",
    corrections: Optional[CorrectionSet],
    array_backends: Iterable[ArrayBackend],
) -> "nt.DTypeLike":
    tmp_dtype: "nt.DTypeLike"
    if corrections is not None and corrections.have_corrections():
        tmp_dtype = np.result_type(np.float32, dtype)
    else:
        tmp_dtype = np.dtype(dtype)
    for udf in udfs:
        tmp_dtype = result_type(
            udf.get_preferred_input_dtype(),
            tmp_dtype,
            *array_backends,
        )
    return tmp_dtype


class TileConverter:
    def __init__(self, tile: DataTile):
        self._tile = tile
        # prime cache with identity
        self._cache = {get_backend(tile.data): tile.data}

    def get(self, backend):
        '''
        Get a converted tile for the specified backend.

        Return from cache if any of them is in the cache,
        otherwise choose cheapest option to convert from availabe cache.
        '''
        res = self._cache.get(backend, None)
        if res is None:
            source, target = cheapest_pair(self._cache.keys(), (backend, ))
            res = for_backend(self._cache[source], target)
            if backend in ND_BACKENDS:
                res = res.reshape(self._tile.shape)
            self._cache[backend] = res
        return res


def _format_plan(plan: ExecutionPlan) -> list[str]:
    res = []
    for k, v in plan.items():
        udf_classes = ",".join([
            str(c.__class__.__name__)
            for c in v
        ])
        res.append(
            f"{str(k)}: {udf_classes}"
        )
    return res


def _execution_plan(
    udfs, ds: Union[DataSet, DataSetMeta], device_class: Optional[DeviceClass] = None,
    available_backends: Iterable[ArrayBackend] = BACKENDS
) -> tuple[ArrayBackend, ExecutionPlan]:
    '''
    Calculate which array format to use for which UDFs
    '''
    remaining = list(udfs)
    execution_plan = OrderedDict()
    # preserve order
    if ds.array_backends is None:
        raise ValueError("Available dataset backends need to be known.")
    ds_backends = tuple(ds.array_backends)
    available_backends = frozenset(available_backends)

    if device_class is None:
        # No preference, for example to prepare running for dataset
        native_backends = available_backends
    elif device_class == 'cuda':
        native_backends = available_backends.intersection(CUDA_BACKENDS)
    elif device_class == 'cpu':
        native_backends = available_backends.intersection(CPU_BACKENDS)
    else:
        raise ValueError(f"Unknown device class{device_class}, allowed are 'cuda' and 'cpu'.")

    aggregate_udf_aversion: dict[ArrayBackend, float] = defaultdict(float)
    for udf in remaining:
        backends = _get_canonical_backends(udf.get_backends())
        for i, b in enumerate(backends):
            aggregate_udf_aversion[b] += i / len(backends)

    backend_popularity = sorted(
        aggregate_udf_aversion.keys(),
        key=lambda b: aggregate_udf_aversion[b]
    )

    def find_popular(remaining, restrict_to, preference_order):
        '''
        Parameters
        ----------

        preference_order: Order of preference for tie breaking if several formats can work for
        the same number of UDFs
        '''
        popular = defaultdict(OrderedDict)
        for udf in remaining:
            for b in _get_canonical_backends(udf.get_backends()):
                if b in restrict_to:
                    # Use dict with value True to make sure UDFs are not counted double
                    popular[b][udf] = True
        most_popular = sorted(popular.keys(), key=lambda b: len(popular[b]), reverse=True)
        if most_popular:
            max_len = len(popular[most_popular[0]])
            most_popular = [b for b in most_popular if len(popular[b]) >= max_len]
            if len(most_popular) > 1:
                for b in preference_order:
                    if b in most_popular:
                        return b, list(popular[b].keys())
            b = most_popular[0]
            return b, list(popular[b].keys())
        else:
            return None, None

    def apply(backend, udfs):
        if backend is not None:
            execution_plan[backend] = udfs
            for udf in udfs:
                remaining.remove(udf)

    # First assign UDFs that work without conversion
    # with a tile format native to dataset and device class
    backend, selected_udfs = find_popular(
        remaining,
        restrict_to=native_backends.intersection(ds_backends),
        preference_order=ds_backends
    )
    if backend is not None:
        apply(backend, selected_udfs)
    # The following covers the common case that the dataset is CPU-only
    # and we require conversion to a CuPy format
    # We should avoid costly conversion dense to sparse or sparse to dense
    # by default when transforming between CPU and CuPy.
    # Assign UDFs with sparse formats if the dataset has any sparse options
    if SPARSE_BACKENDS.intersection(ds_backends):
        backend, selected_udfs = find_popular(
            remaining,
            restrict_to=native_backends.intersection(SPARSE_BACKENDS),
            preference_order=backend_popularity
        )
        if backend is not None:
            apply(backend, selected_udfs)
    # Assign a dense native format if the dataset has dense options
    if DENSE_BACKENDS.intersection(ds_backends):
        backend, selected_udfs = find_popular(
            remaining,
            restrict_to=native_backends.intersection(DENSE_BACKENDS),
            preference_order=backend_popularity
        )
        if backend is not None:
            apply(backend, selected_udfs)
    # Then assign UDFs that work with an array format that matches the devie class
    # That means arrays will be converted to allow execution on GPU or CPU even if the
    # dataset can't deliver that natively and it requires conversion between sparse and dense.
    # That is the case for the dataset default NumPy and UDFs that require sparse input.
    while True:
        backend, selected_udfs = find_popular(
            remaining,
            restrict_to=native_backends,
            preference_order=backend_popularity
        )
        if backend is not None:
            apply(backend, selected_udfs)
        else:
            break
    # From here we have the "CPU on GPU fallback". A GPU on CPU fallback
    # is attempted as well, may trigger an error downstream.
    # Assign remaining UDFs that work with an array backend native to the dataset
    backend, selected_udfs = find_popular(
        remaining,
        restrict_to=available_backends.intersection(ds_backends),
        preference_order=ds_backends
    )
    if backend:
        apply(backend, selected_udfs)
    # Finally assign UDFs to any array backend even if they don't match the
    # dataset or device class.
    while True:
        backend, selected_udfs = find_popular(
            remaining,
            restrict_to=available_backends,
            preference_order=backend_popularity
        )
        if backend is not None:
            apply(backend, selected_udfs)
        else:
            break

    if remaining:
        for r in remaining:
            for b in _get_canonical_backends(r.get_backends()):
                if b not in BACKENDS:
                    raise ValueError(f"UDF {r} returned invalid backend {b}.")
        raise RuntimeError(f"Could not find backends for remaining {remaining}.")

    # Optimize the execution plan
    reordered_plan = OrderedDict()
    available = []
    required = list(execution_plan.keys())
    # First, determine the start point
    ds_backend, right = cheapest_pair(ds_backends, required)
    reordered_plan[right] = execution_plan[right]
    available.append(ds_backend)
    available.append(right)
    required.remove(right)
    # Now assign the remaining ones
    while required:
        _, right = cheapest_pair(available, required)
        reordered_plan[right] = execution_plan[right]
        required.remove(right)
    # Now we should have covered all

    trace.get_current_span().add_event(
        'execution_plan', {
            'ds_backend': ds_backend,
            'execution_plan': _format_plan(reordered_plan),
        }
    )
    return ds_backend, reordered_plan


class UDFMeta:
    """
    UDF metadata. Makes all relevant metadata accessible to the UDF. Can be different
    for each task/partition.

    .. versionchanged:: 0.4.0
        Added distinction of dataset_dtype and input_dtype

    .. versionchanged:: 0.6.0
        Information on compute backend, corrections, coordinates and tiling scheme added

    .. versionchanged:: 0.9.0
        :code:`tiling_scheme_idx` and `sig_slice` added

    .. versionchanged:: 0.11.0
        :code:`array_backend` added
    """

    def __init__(
        self,
        partition_slice: Optional[Slice],
        dataset_shape: Shape,
        roi: Optional[np.ndarray],
        dataset_dtype: "nt.DTypeLike",
        input_dtype: "nt.DTypeLike",
        tiling_scheme: Optional[TilingScheme] = None,
        tiling_index: int = 0,
        corrections: Optional[CorrectionSet] = None,
        device_class: Optional[DeviceClass] = None,
        threads_per_worker: Optional[int] = None,
        array_backend: Optional[ArrayBackend] = None,
        valid_nav_mask: Optional[np.ndarray] = None,
    ):
        self._partition_slice = partition_slice
        self._dataset_shape = dataset_shape
        self._dataset_dtype = dataset_dtype
        self._input_dtype = input_dtype
        self._tiling_scheme = tiling_scheme
        self._tiling_index = tiling_index
        if device_class is None:
            device_class = 'cpu'
        self._device_class = device_class
        if roi is not None:
            roi = roi.reshape(tuple(dataset_shape.nav))
        self._roi = roi
        self._slice: Optional[Slice] = None
        self._cached_coordinates: Optional[np.ndarray] = None
        if corrections is None:
            corrections = CorrectionSet()
        self._corrections = corrections
        self._threads_per_worker = threads_per_worker
        self._array_backend = array_backend
        self._valid_nav_mask = valid_nav_mask

        if valid_nav_mask is not None and roi is not None:
            assert valid_nav_mask.shape[0] == count_nonzero(roi)

    @property
    def slice(self) -> Optional[Slice]:
        """
        Slice : A :class:`~libertem.common.slice.Slice` instance that describes the location
                within the dataset with navigation dimension flattened and reduced to the ROI.
        """
        return self._slice

    @slice.setter
    def slice(self, new_slice: Slice) -> None:
        self._slice = new_slice

    @property
    def partition_shape(self) -> Shape:
        """
        Shape : The shape of the partition this UDF currently works on.
                If a ROI was applied, the shape will be modified accordingly.
        """
        if self._partition_slice is None:
            raise ValueError("cannot get partition_shape if partition_slice is None")
        return self._partition_slice.shape

    @property
    def dataset_shape(self) -> Shape:
        """
        Shape : The original shape of the whole dataset, not influenced by the ROI
        """
        return self._dataset_shape

    @property
    def tiling_scheme(self) -> Optional[TilingScheme]:
        """
        TilingScheme : the tiling scheme that was negotiated

        .. versionadded:: 0.6.0
        """
        return self._tiling_scheme

    @property
    def tiling_scheme_idx(self) -> int:
        '''
        Index of the current tile in :attr:`tiling_scheme`.
        '''
        return self._tiling_index

    @tiling_scheme_idx.setter
    def tiling_scheme_idx(self, new_idx: int) -> None:
        self._tiling_index = new_idx

    @property
    def sig_slice(self) -> Slice:
        '''
        Signal slice of the current tile.

        Since all tiles follow the same tiling scheme, this avoids repeatedly
        calculating the signal part of :attr:`slice`. Instead, the
        appropriate slice from the tiling scheme can be re-used.
        '''
        assert self._tiling_scheme is not None
        return self._tiling_scheme[self._tiling_index]

    @property
    def roi(self) -> Optional[np.ndarray]:
        """
        numpy.ndarray : Boolean array which limits the elements the UDF is working on.
                     Has a shape of :attr:`dataset_shape.nav`.
        """
        return self._roi

    @property
    def dataset_dtype(self) -> "nt.DTypeLike":
        """
        numpy.dtype : Native dtype of the dataset
        """
        return self._dataset_dtype

    @property
    def input_dtype(self) -> "nt.DTypeLike":
        """
        numpy.dtype : dtype of the data that will be passed to the UDF

        This is determined from the dataset's native dtype and
        :meth:`UDF.get_preferred_input_dtype` using :meth:`numpy.result_type`

        .. versionadded:: 0.4.0
        """
        return self._input_dtype

    @property
    def corrections(self) -> CorrectionSet:
        """
        CorrectionSet : correction data that is available, either from the dataset
        or specified by the user

        .. versionadded:: 0.6.0
        """
        return self._corrections

    @property
    def device_class(self) -> str:
        '''
        Which device class is used.

        The back-end library can be accessed as :attr:`libertem.udf.base.UDF.xp`.
        This additional string information is used since that way the back-end can be probed without
        importing them all and testing them against :attr:`libertem.udf.base.UDF.xp`.

        Current values are :code:`cpu` (default) or :code:`cuda`.

        .. versionadded:: 0.6.0
        '''
        return self._device_class

    @property
    def coordinates(self) -> np.ndarray:
        """
        np.ndarray : Array of coordinates that correspond to the frames in the actual
        navigation space which are part of the current tile or partition.

        .. versionadded:: 0.6.0
        """
        assert self._slice is not None
        assert self._partition_slice is not None
        if self._cached_coordinates is None:
            self._cached_coordinates = get_coordinates(
                self._partition_slice,
                self._dataset_shape,
                self._roi
            )
        shifted_slice = self._slice.shift(self._partition_slice).get(nav_only=True)
        return self._cached_coordinates[shifted_slice]

    @property
    def threads_per_worker(self) -> Optional[int]:
        """
        int or None : number of threads that a UDF is allowed to use in the `process_*` method.
                      For Numba, pyfftw, Torch, NumPy and SciPy (OMP, MKL, OpenBLAS), this limit
                      is set automatically; this property can be used for other cases, like manually
                      creating thread pools or setting limits for unsupported modules.
                      :code:`None` means no limit is set, and the UDF can use any number of threads
                      it deems necessary (should be limited to system limits, of course).

        Note
        ----

        .. versionchanged:: 0.8.0
            Since discovery of loaded libraries can be slow with :mod:`threadpoolctl`
            (:issue:`1117`), they are cached now. In case an UDF triggers loading of a new
            library or instance of a library that is supported by :mod:`threadpoolctl`,
            it will only be discovered in the first run on a :class:`~libertem.api.Context`.
            The threading settings for such other libraries or instances can therefore depend on the
            execution order. In such cases the thread count for affected libraries should be
            set in the UDF based on :code:`threads_per_worker`. Numba, pyfftw, Torch, NumPy
            and SciPy should not be affected since they are loaded before the first discovery.

        See also: :func:`libertem.common.threading.set_num_threads`

        .. versionadded:: 0.7.0
        """
        return self._threads_per_worker

    @property
    def array_backend(self) -> Optional[ArrayBackend]:
        """
        Array backend, one of the constants defined in :code:`BACKEND_*`
        constants in :class:`libertem.udf.base.UDF`, or None if not known at
        that time.

        .. versionadded:: 0.11.0
        """
        return self._array_backend

    def get_valid_nav_mask(self, full_nav: bool = False) -> Optional[np.ndarray]:
        """
        Return a mask of the already computed nav positions, as flattened
        1D array.

        Only available in :meth:`~libertem.udf.base.UDF.merge` and
        :meth:`~libertem.udf.base.UDF.get_results`.

        In case of :meth:`~libertem.udf.base.UDF.merge`, the mask does not yet
        contain the positions of the data that will be merged into the result,
        but only those positions that have already been merged into result.

        NOTE: these positions may include dropped frames or missing data.

        Parameters
        ----------
        full_nav
            If a `roi` is applied, still include all elements of the navigation
            space in the mask. If this is not set, the mask is compressed to the
            positions selected in the `roi`.
        """
        if self._valid_nav_mask is None:
            return None
        if full_nav and self._roi is not None:
            ds_shape = self.dataset_shape.nav.to_tuple()
            full_mask = np.zeros(ds_shape, dtype=bool).reshape((-1,))
            full_mask[self._roi.reshape((-1,))] = self._valid_nav_mask
            return full_mask
        else:
            return self._valid_nav_mask

    def set_valid_nav_mask(self, new_valid_nav_mask: Optional[np.ndarray]):
        self._valid_nav_mask = new_valid_nav_mask


class MergeAttrMapping:
    def __init__(self, dict_input: dict[str, np.ndarray]):
        self._dict: dict[str, np.ndarray] = dict_input

    def __iter__(self) -> Iterator[str]:
        return iter(self._dict)

    def __contains__(self, k: str) -> bool:
        return k in self._dict

    def __getattr__(self, k: str) -> np.ndarray:
        return self._dict[k]

    def __setattr__(self, k: str, v: np.ndarray) -> None:
        if k in ['_dict']:
            super().__setattr__(k, v)
        else:
            self._dict[k][:] = v

    def __getitem__(self, k: str) -> np.ndarray:
        warnings.warn(
            "dict-like access is discouraged, as it can be "
            "confusing vs. using attribute access",
            UseDiscouragedWarning,
            stacklevel=2,
        )
        return self._dict[k]


T = TypeVar('T')


class UDFData:
    '''
    Container for result buffers, return value from running UDFs
    '''

    def __init__(self, data: dict[str, BufferWrapper]):
        self._data = data
        self._views: dict[str, np.ndarray] = {}

    def __repr__(self) -> str:
        return "<UDFData: %r>" % (
            self._data
        )

    def __getattr__(self, k: str) -> Union[np.ndarray, BufferWrapper]:
        if k.startswith("_"):
            raise AttributeError("no such attribute: %s" % k)
        try:
            return self._get_view_or_data(k)
        except KeyError as e:
            raise AttributeError(str(e))

    def get_buffer(self, name: str) -> BufferWrapper:
        """
        Return the `BufferWrapper` for buffer `name`

        .. versionadded:: 0.7.0
        """
        return self._data[name]

    def set_buffer(self, name: str, buffer: BufferWrapper) -> None:
        """
        Replace or set the `BufferWrapper` for buffer `name`
        """
        assert isinstance(buffer, BufferWrapper)
        self._data[name] = buffer

    def get(
        self, k: str, default: Optional[T] = None
    ) -> Optional[Union[T, np.ndarray, BufferWrapper]]:
        try:
            return self.__getattr__(k)
        except (KeyError, AttributeError):
            return default

    def __setattr__(self, k: str, v: "nt.ArrayLike") -> None:
        if not k.startswith("_"):
            # convert UDFData.some_attr = something to array slice assignment
            getattr(self, k)[:] = v
        else:
            super().__setattr__(k, v)

    def _get_view_or_data(self, k: str) -> Union[np.ndarray, BufferWrapper]:
        if k in self._views:
            return self._views[k]
        res = self._data[k]
        if isinstance(res, BufferWrapper) and res.raw_data is not None:
            return res.raw_data
        return res

    def __getitem__(self, k: str) -> BufferWrapper:
        warnings.warn(
            "dict-like access is discouraged, as it can be "
            "confusing vs. using attribute access. Please use `get_buffer` instead, "
            "if you really need the `BufferWrapper` and not the current view",
            UseDiscouragedWarning,
            stacklevel=2,
        )
        return self._data[k]

    def __contains__(self, k: str) -> bool:
        return k in self._data

    def items(self):
        return self._data.items()

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def as_dict(self) -> dict[str, BufferWrapper]:
        return dict(self.items())

    def get_proxy(self) -> MergeAttrMapping:
        return MergeAttrMapping({
            k: (self._views[k] if k in self._views else self._data[k].raw_data)
            for k, v in self.items()
            if v and v.has_data()
        })

    def _get_buffers(
        self, filter_allocated: bool = False
    ) -> Generator[tuple[str, BufferWrapper], None, None]:
        for k, buf in self._data.items():
            if not hasattr(buf, 'has_data') or (buf.has_data() and filter_allocated):
                continue
            yield k, buf

    def allocate_for_part(self, partition: Partition, roi: Optional[np.ndarray], lib=None) -> None:
        """
        allocate all BufferWrapper instances in this namespace.
        for pre-allocated buffers (i.e. aux data), only set shape and roi
        """
        for k, buf in self._get_buffers():
            buf.set_shape_partition(partition, roi)
        for k, buf in self._get_buffers(filter_allocated=True):
            buf.allocate(lib=lib)

    def allocate_for_full(self, dataset: DataSet, roi: Optional[np.ndarray]) -> None:
        ds_partitions = [*dataset.get_partitions()]
        for k, buf in self._get_buffers():
            buf.set_shape_ds(dataset.shape, roi)
            buf.add_partitions(ds_partitions)
        for k, buf in self._get_buffers(filter_allocated=True):
            buf.allocate()

    def set_view_for_dataset(self, dataset: DataSet) -> None:
        for k, buf in self._get_buffers():
            self._views[k] = buf.get_view_for_dataset(dataset)

    def set_view_for_partition(self, partition: Partition) -> None:
        for k, buf in self._get_buffers():
            self._views[k] = buf.get_view_for_partition(partition)

    def set_view_for_tile(self, partition: Partition, tile: DataTile) -> None:
        for k, buf in self._get_buffers():
            self._views[k] = buf.get_view_for_tile(partition, tile)

    def set_contiguous_view_for_tile(self, partition: Partition, tile: DataTile) -> None:
        # .. versionadded:: 0.5.0
        for k, buf in self._get_buffers():
            self._views[k] = buf.get_contiguous_view_for_tile(partition, tile)

    def flush(self, debug: bool = False) -> None:
        # .. versionadded:: 0.5.0
        for k, buf in self._get_buffers():
            buf.flush(debug=debug)

    def export(self) -> None:
        # .. versionadded:: 0.6.0
        for k, buf in self._get_buffers():
            buf.export()

    def set_view_for_frame(self, partition: Partition, tile: DataTile, frame_idx: int) -> None:
        for k, buf in self._get_buffers():
            self._views[k] = buf.get_view_for_frame(partition, tile, frame_idx)

    def clear_views(self) -> None:
        self._views = {}


class UDFKwargsWrapper(UDFData):
    """
    Wrapper for UDF kwargs, used for slicing/viewing AUX data
    """
    def __init__(self, data: UDFKwargs):
        super().__init__(data)
        self._data = data
        self._views = {}

    def _get_buffers(
        self, filter_allocated: bool = False
    ) -> Generator[tuple[str, AuxBufferWrapper], None, None]:
        for k, buf in self._data.items():
            if isinstance(buf, AuxBufferWrapper):
                if buf.has_data() and filter_allocated:
                    continue
                yield k, buf

    def new_for_partition(self, partition: Partition, roi: np.ndarray):
        for k, buf in self._get_buffers():
            self._data[k] = buf.new_for_partition(partition, roi)


@runtime_checkable
class UDFFrameMixin(Protocol):
    '''
    Implement :code:`process_frame` for per-frame processing.
    '''

    def process_frame(self, frame: np.ndarray):
        """
        Implement this method to process the data on a frame-by-frame manner.

        Data available in this method:

        - `self.params`    - the parameters of this UDF
        - `self.task_data` - task data created by `get_task_data`
        - `self.results`   - the result buffer instances
        - `self.meta`      - meta data about the current operation and data set

        Parameters
        ----------
        frame : numpy.ndarray or cupy.ndarray
            A single frame or signal element from the dataset.
            The shape is the same as `dataset.shape.sig`. In case of pixelated
            STEM / scanning diffraction data this is 2D, for spectra 1D etc.
        """
        raise NotImplementedError()


@runtime_checkable
class UDFTileMixin(Protocol):
    '''
    Implement :code:`process_tile` for per-tile processing.
    '''

    def process_tile(self, tile: np.ndarray):
        """
        Implement this method to process the data in a tiled manner.

        Data available in this method:

        - `self.params`    - the parameters of this UDF
        - `self.task_data` - task data created by `get_task_data`
        - `self.results`   - the result buffer instances
        - `self.meta`      - meta data about the current operation and data set

        Parameters
        ----------
        tile : numpy.ndarray or cupy.ndarray
            A small number N of frames or signal elements from the dataset.
            The shape is (N,) + `dataset.shape.sig`. In case of pixelated
            STEM / scanning diffraction data this is 3D, for spectra 2D etc.
        """
        raise NotImplementedError()


@runtime_checkable
class UDFPartitionMixin(Protocol):
    '''
    Implement :code:`process_partition` for per-partition processing.
    '''

    def process_partition(self, partition: np.ndarray) -> None:
        """
        Implement this method to process the data partitioned into large
        (100s of MiB) partitions.

        Data available in this method:

        - `self.params`    - the parameters of this UDF
        - `self.task_data` - task data created by `get_task_data`
        - `self.results`   - the result buffer instances
        - `self.meta`      - meta data about the current operation and data set

        Note
        ----
        Only use this method if you know what you are doing; especially if
        you are running a processing pipeline with multiple steps, or multiple
        processing pipelines at the same time, performance may be adversely
        impacted.

        Parameters
        ----------
        partition : numpy.ndarray or cupy.ndarray
            A large number N of frames or signal elements from the dataset.
            The shape is (N,) + `dataset.shape.sig`. In case of pixelated
            STEM / scanning diffraction data this is 3D, for spectra 2D etc.
        """
        raise NotImplementedError()


@runtime_checkable
class UDFPreprocessMixin(Protocol):
    '''
    Implement :code:`preprocess` to initialize the result buffers of a partition on the worker
    before the partition data is processed.

    .. versionadded:: 0.3.0
    '''

    def preprocess(self) -> None:
        """
        Implement this method to preprocess the result data for a partition.

        This can be useful to initialize arrays of
        :code:`dtype='object'` with the correct container types, for example.

        Data available in this method:

        - `self.params`    - the parameters of this UDF
        - `self.task_data` - task data created by `get_task_data`
        - `self.results`   - the result buffer instances
        """
        raise NotImplementedError()


@runtime_checkable
class UDFPostprocessMixin(Protocol):
    '''
    Implement :code:`postprocess` to modify the resulf buffers of a partition on the worker
    after the partition data has been completely processed, but before it is returned to the
    main node for the final merging step.
    '''

    def postprocess(self) -> None:
        """
        Implement this method to postprocess the result data for a partition.

        This can be useful in combination with process_tile() to implement
        a postprocessing step that requires the reduced results for whole frames.

        Data available in this method:

        - `self.params`    - the parameters of this UDF
        - `self.task_data` - task data created by `get_task_data`
        - `self.results`   - the result buffer instances
        """
        raise NotImplementedError()


@runtime_checkable
class UDFMergeAllMixin(Protocol):
    def merge_all(
                self, ordered_results: 'OrderedDict[Slice, MergeAttrMapping]'
            ) -> Mapping[str, 'nt.ArrayLike']:
        """
        Combine stack of ordered partial results :code:`ordered_results` to form complete result.

        Combining can be more efficient than direct merging into a result buffer
        for cases where the results are not NumPy arrays.
        Currently this is only applicable for the
        :class:`libertem.executor.delayed.DelayedJobExecutor`
        where it provides an efficient pathway to construct Dask arrays from delayed UDF results.

        The input and the returned arrays are in flattened navigation dimension with ROI applied.

        Data available in this method:

        - :code:`self.params` - the parameters of this UDF

        For UDFs with only :code:`kind='nav'` result buffers a default implementation
        is used automatically.

        Parameters
        ----------

        ordered_results
            Ordered dict mapping partition slice to UDF partial result

        Returns
        -------

        dict[buffername] -> array_like
            Dictionary mapping result buffer name to buffer content

        Note
        ----
        This function is running on the main node, which means `self.results`
        and `self.task_data` are not available.
        """
        raise NotImplementedError()


def _default_merge_all(udf, ordered_results: 'OrderedDict[Slice, MergeAttrMapping]'):
    if udf.requires_custom_merge_all:
        raise NotImplementedError(
            "Default merging only works for kind='nav' buffers. "
            "Please implement a suitable custom merge_all function."
        )
    result_chunks = defaultdict(list)
    for b in ordered_results.values():
        for key in b:
            result_chunks[key].append(getattr(b, key))

    result = {
        # checking above assures that we only have kind='nav'
        # where concatenation is the correct method.
        k: np.concatenate(val)
        for k, val in result_chunks.items()
    }
    return result


class UDFBase(UDFProtocol):
    '''
    Base class for UDFs with helper functions.
    '''
    def __init__(self, *args, **kwargs) -> None:
        # this should not execute - it is mostly for getting the typing right:
        self.params = UDFKwargsWrapper({})
        self.results = UDFData({})

    def _update_parameters_patch(self, kwargs_patch):
        """
        Update the parameters _for this UDF instance_ by applying the patch in
        `kwargs_patch`, which is a dict containing the key/value combinations to
        update.

        For properly supporting AUX data (i.e. `BufferWrapper`-typed parameters),
        make sure to set the correct views afterwards.

        :meta private:
        """
        new_kwargs = {}
        new_kwargs.update(self._kwargs)
        new_kwargs.update(kwargs_patch)
        self.params = UDFKwargsWrapper(new_kwargs)

    def get_task_data(self) -> dict[str, Any]:
        raise NotImplementedError()

    def get_result_buffers(self) -> dict[str, BufferWrapper]:
        raise NotImplementedError()

    def allocate_for_part(self, partition: Partition, roi: Optional[np.ndarray]) -> None:
        for ns in [self.results]:
            ns.allocate_for_part(partition, roi, lib=self.xp)

    def allocate_for_full(self, dataset: DataSet, roi: Optional[np.ndarray]) -> None:
        for ns in [self.params, self.results]:
            ns.allocate_for_full(dataset, roi)

    def set_views_for_dataset(self, dataset: DataSet) -> None:
        for ns in [self.params]:
            ns.set_view_for_dataset(dataset)

    def set_views_for_partition(self, partition: Partition) -> None:
        for ns in [self.params, self.results]:
            ns.set_view_for_partition(partition)

    def set_views_for_tile(self, partition: Partition, tile: DataTile) -> None:
        for ns in [self.params, self.results]:
            ns.set_view_for_tile(partition, tile)

    def set_contiguous_views_for_tile(self, partition: Partition, tile: DataTile) -> None:
        # .. versionadded:: 0.5.0
        for ns in [self.params, self.results]:
            ns.set_contiguous_view_for_tile(partition, tile)

    def flush(self, debug: bool = False) -> None:
        # .. versionadded:: 0.5.0
        for ns in [self.params, self.results]:
            ns.flush(debug=debug)

    def set_views_for_frame(self, partition: Partition, tile: DataTile, frame_idx: int):
        for ns in [self.params, self.results]:
            ns.set_view_for_frame(partition, tile, frame_idx)

    def clear_views(self) -> None:
        for ns in [self.params, self.results]:
            ns.clear_views()

    def init_task_data(self) -> None:
        self.task_data = UDFData(self.get_task_data())

    def init_result_buffers(self, executor=None) -> None:
        """
        Create the UDFData instance containing BufferWrappers
        for results of this UDF. Same method used for
        Dataset- and Partition-sized buffers.

        The executor argument is provided on the main node
        only to allow modification of Dataset-sized buffers
        buffers into more specialized forms.

        #TODO use another mechanism to distinguish between
        dataset and parititon-sized buffers when initializing,
        such that we can modify buffers in both cases.

        :meta private:
        """
        self.results = UDFData(self.get_result_buffers())
        if executor is not None:
            for name, buffer in self.results.items():
                new_buffer = executor.modify_buffer_type(buffer)
                self.results.set_buffer(name, new_buffer)

    def export_results(self) -> None:
        # .. versionadded:: 0.6.0.dev0
        self.results.export()

    def set_meta(self, meta: UDFMeta) -> None:
        self.meta = meta

    def set_slice(self, slice_: Slice) -> None:
        self.meta.slice = slice_

    def set_tile_idx(self, idx: int) -> None:
        self.meta.tiling_scheme_idx = idx

    def set_backend(self, backend: ArrayBackend) -> None:
        assert backend in _get_canonical_backends(self.get_backends())
        self._backend = backend

    def get_backends(self) -> BackendSpec:
        raise NotImplementedError()  # see impl in UDF.get_backends

    @property
    def xp(self):
        '''
        Compute back-end library to use.

        Generally, use :code:`self.xp` instead of :code:`np` to use NumPy or CuPy transparently

        .. versionadded:: 0.6.0
        '''
        # Implemented as property and not variable to avoid pickling issues
        # tests/udf/test_simple_udf.py::test_udf_pickle
        if self._backend in CPU_BACKENDS or self._backend == CUDA:
            return np
        elif self._backend in CUPY_BACKENDS:
            # Re-importing should be fast, right?
            # Importing only here to avoid superfluous import
            import cupy
            # mocking for testing without actual CUDA device
            # import numpy as cupy
            return cupy
        else:
            raise ValueError(f"Backend can be {BACKENDS}, got {self._backend}")

    def get_method(self) -> UDFMethod:
        """
        Return a member of the :attr:`libertem.udf.base.UDF.UDF_METHOD`
        enum to indicate which processing method to use during :code:`run_udf`.

        By default, this method returns based on the process method
        precedence of tile > frame > partition.

        The return is checked at runtime to ensure the requested method
        is impemented on the UDF itself.

        .. versionchanged:: 0.13.0
            :meth:`~libertem.udf.base.UDF.get_method()` was exposed to enable
            choice of processing method at runtime

        Returns
        -------

        UDFMethod
            A member of the enum :attr:`libertem.udf.base.UDF.UDF_METHOD`
        """
        if isinstance(self, UDFTileMixin):
            return UDFMethod.TILE
        elif isinstance(self, UDFFrameMixin):
            return UDFMethod.FRAME
        elif isinstance(self, UDFPartitionMixin):
            return UDFMethod.PARTITION
        else:
            raise TypeError("UDF should implement one of the `process_*` methods")

    def _check_results(self, decl, arr, name):
        """
        Check results in `arr` for buffer `name` for consistency with the
        declaration in `decl`.

        1) All returned buffers need to be declared
        2) Private buffers can't be returned from `get_results`
        3) The `dtype` of `arr` needs to be compatible with the declared `dtype`
        """
        if name not in decl:
            raise UDFException(
                "buffer '%s' is not declared in `get_result_buffers` "
                "(hint: `self.buffer(..., use='result_only')`" % name
            )
        buf_decl = decl[name]
        if buf_decl.use == "private":
            raise UDFException("Don't return `use='private'` buffers from `get_results`")
        if np.dtype(arr.dtype).kind != np.dtype(buf_decl.dtype).kind:
            raise UDFException(
                "the returned ndarray '%s' has a different dtype kind (%s) "
                "than declared (%s)" % (
                    name, arr.dtype, buf_decl.dtype
                )
            )

    def get_results(self) -> dict[str, np.ndarray]:
        raise NotImplementedError()

    _default_merge_all = UDFMergeAllMixin.merge_all

    def _do_merge_all(self, ordered_results: 'OrderedDict[Slice, MergeAttrMapping]'):
        if isinstance(self, UDFMergeAllMixin):
            results_tmp = self.merge_all(ordered_results)
        else:
            # This will raise NotImplemementedError if preconditions for default merge
            # are not met
            results_tmp = _default_merge_all(self, ordered_results)

        if not set(results_tmp.keys()).issubset(set(self.results.keys())):
            raise ValueError(f'Returned result names from merge_all ({[*results_tmp.keys()]}) '
                            'are not contained within declared result buffer names '
                            f'({[*self.results.keys()]})')

        for key, value in results_tmp.items():
            # This SHOULD throw errors if sth doesn't match up about
            # buffer name, shape or dtype
            self.results.get_buffer(key).replace_array(value)

    def _do_get_results(self) -> Mapping[str, BufferWrapper]:
        results_tmp = self.get_results()
        decl = self.get_result_buffers()

        # include any results that were not explicitly included, but have non-private `use`:
        results_tmp.update({
            k: getattr(self.results, k)
            for k, v in decl.items()
            if k not in results_tmp and v.use is None
        })

        results_buffer_cls = {
            k: self.results.get_buffer(k).result_buffer_type()
            for k in results_tmp.keys()
        }

        # wrap numpy results into `ResultBuffer`s:
        results = {}
        for name, arr in results_tmp.items():
            mask = None
            if isinstance(arr, ArrayWithMask):
                mask = arr.mask
                arr = arr.arr
            self._check_results(decl, arr, name)
            buf_decl = decl[name]
            if mask is None:
                valid_mask = self.meta.get_valid_nav_mask()
                assert valid_mask is not None
                mask = buf_decl.make_default_mask(
                    valid_nav_mask=valid_mask,
                    dataset_shape=self.meta.dataset_shape,
                    roi=self.meta.roi,
                )
            buf = results_buffer_cls[name](
                kind=buf_decl.kind, extra_shape=buf_decl.extra_shape,
                dtype=buf_decl.dtype,
                data=arr,
            )
            buf.valid_mask = mask
            buf.set_shape_ds(self.meta.dataset_shape, self.meta.roi)
            results[name] = buf
        return results


class UDF(UDFBase):
    """
    The main user-defined functions interface. You can implement your functionality
    by overriding methods on this class.

    If you override `__init__`, please take care,
    as it is called multiple times during evaluation of a UDF. You can handle
    some pre-conditioning of parameters, but you also have to accept the results
    as input again.

    Arguments passed as `**kwargs` will be automatically available on `self.params`
    when running the UDF.

    Example
    -------

    >>> class MyUDF(UDF):
    ...     def __init__(self, param1, param2="def2", **kwargs):
    ...         param1 = int(param1)
    ...         if "param3" not in kwargs:
    ...             raise TypeError("missing argument param3")
    ...         super().__init__(param1=param1, param2=param2, **kwargs)

    Parameters
    ----------
    kwargs
        Input parameters. They are scattered to the worker processes and
        available as `self.params` from here on.

        Values can be `BufferWrapper` instances, which, when accessed via
        `self.params.the_key_here`, will automatically return a view corresponding
        to the current unit of data (frame, tile, partition).
    """

    def __init__(self, **kwargs: UDFKwarg) -> None:
        self._backend = 'numpy'  # default so that self.xp can always be used
        self._kwargs = kwargs
        self.params = UDFKwargsWrapper(kwargs)
        self.task_data = UDFData({})
        self.results = UDFData({})
        self._requires_custom_merge = None
        self._requires_custom_merge_all = None

    def copy(self) -> "UDF":
        return self.__class__(**self._kwargs)

    @classmethod
    def new_for_partition(
        cls,
        kwargs: dict[str, Any],
        partition: Partition,
        roi: np.ndarray,
    ) -> "UDF":
        new_instance = cls(**kwargs)
        new_instance.params.new_for_partition(partition, roi)
        return new_instance

    def copy_for_partition(self, partition: Partition, roi: np.ndarray) -> "UDF":
        """
        create a copy of the UDF, specifically slicing aux data to the
        specified pratition and roi
        """
        return self.__class__.new_for_partition(self._kwargs, partition, roi)

    def get_task_data(self) -> dict[str, Any]:
        """
        Initialize per-task data.

        Per-task data can be mutable. Override this function
        to allocate temporary buffers, or to initialize
        system resources.

        If you want to distribute static data, use
        parameters instead.

        Data available in this method:

        - `self.params` - the input parameters of this UDF
        - `self.meta` - relevant metadata, see :class:`UDFMeta` documentation.

        Returns
        -------
        dict
            Flat dict with string keys. Keys should
            be valid python identifiers, which allows
            access via `self.task_data.the_key_here`.
        """
        return {}

    def get_result_buffers(self) -> dict[str, BufferWrapper]:
        """
        Return result buffer declaration.

        Values of the returned dict should be :class:`~libertem.common.buffers.BufferWrapper`
        instances, which, when accessed via :code:`self.results.key`,
        will automatically return a view corresponding to the
        current unit of data (frame, tile, partition).

        The values also need to be serializable via pickle.

        Data available in this method:

        - :code:`self.params` - the parameters of this UDF
        - :code:`self.meta` - relevant metadata, see :class:`UDFMeta` documentation.
          Please note that partition metadata will not be set when this method is
          executed on the head node.

        Returns
        -------
        dict
            Flat dict with string keys. Keys should
            be valid python identifiers, which allows
            access via `self.results.the_key_here`.
        """
        raise NotImplementedError()

    @property
    def requires_custom_merge(self):
        """
        Determine if buffers with :code:`kind != 'nav'` that are not :code:`use
        != 'result_only'` are present where the default merge doesn't work

        .. versionadded:: 0.5.0

        .. versionchanged:: 0.12.0
            Ignore :code:`use != 'result_only'` buffer since they are not present
            in the merge function
        """
        if self._requires_custom_merge is None:
            buffers = self.get_result_buffers()
            self._requires_custom_merge = any(
                buffer.kind != 'nav' and buffer.use != 'result_only' for buffer in buffers.values()
            )
        return self._requires_custom_merge

    @property
    def requires_custom_merge_all(self):
        """
        Determine if buffers with :code:`kind != 'nav'` are present where
        the default merge doesn't work

        .. versionadded:: 0.12.0
        """
        if self._requires_custom_merge_all is None:
            buffers = self.get_result_buffers()
            self._requires_custom_merge_all = any(
                buffer.kind != 'nav' for buffer in buffers.values()
            )
        return self._requires_custom_merge_all

    def merge(self, dest: MergeAttrMapping, src: MergeAttrMapping):
        """
        Merge a partial result `src` into the current global result `dest`.

        Data available in this method:

        - `self.params` - the parameters of this UDF

        Parameters
        ----------

        dest
            global results; you can access the ndarrays for each
            buffer name (from `get_result_buffers`) by attribute access
            (:code:`dest.your_buffer_name`)

        src
            results for a partition; you can access the ndarrays for each
            buffer name (from `get_result_buffers`) by attribute access
            (:code:`src.your_buffer_name`)

        Note
        ----
        This function is running on the main node, which means `self.results`
        and `self.task_data` are not available.
        """
        if self.requires_custom_merge:
            raise NotImplementedError(
                "Default merging only works for kind='nav' buffers. "
                "Please implement a suitable custom merge function."
            )
        for k in dest:
            check_cast(getattr(src, k), getattr(dest, k))
            getattr(dest, k)[:] = getattr(src, k)

    def get_results(self) -> dict[str, np.ndarray]:
        """
        Get results, allowing a postprocessing step on the main node after
        a result has been merged. See also: :class:`UDFPostprocessMixin`.

        .. versionadded:: 0.7.0

        Note
        ----
        You should return all values as numpy arrays, they will be wrapped
        in `BufferWrapper` instances before they are returned to the user.

        See the :ref:`udf final post processing` section in the documentation for
        details and examples.

        Returns
        -------

        results : dict
            A `dict` containing the final post-processed results.

        """
        for k in self.results.keys():
            buf = self.results.get_buffer(k)
            if buf.use == "result_only":
                raise UDFException(
                    "don't know how to set use='result_only' buffer '%s';"
                    " please implement `get_results`" % k
                )
        decls = self.get_result_buffers()
        return {
            k: getattr(self.results, k)
            for k in self.results.keys()
            if decls[k].use != "private"
        }

    def get_preferred_input_dtype(self) -> "nt.DTypeLike":
        '''
        Override this method to specify the preferred input dtype of the UDF.

        The default is :code:`float32` since most numerical processing tasks
        perform best with this dtype, namely dot products.

        The back-end uses this preferred input dtype in combination with the
        dataset`s native dtype to determine the input dtype using
        :func:`numpy.result_type`. That means :code:`float` data in a dataset
        switches the dtype to :code:`float` even if this method returns an
        :code:`int` dtype. :code:`int32` or wider input data would switch from
        :code:`float32` to :code:`float64`, and complex data in the dataset will
        switch the input dtype kind to :code:`complex`, following the NumPy
        casting rules.

        In case your UDF only works with specific input dtypes, it should throw
        an error or warning if incompatible dtypes are used, and/or implement a
        meaningful conversion in your UDF's :code:`process_<...>` routine.

        If you prefer to always use the dataset's native dtype instead of
        floats, you can override this method to return
        :attr:`UDF.USE_NATIVE_DTYPE`, which is currently identical to
        :code:`numpy.bool` and behaves as a neutral element in
        :func:`numpy.result_type`.

        .. versionadded:: 0.4.0
        '''
        return np.float32

    def get_tiling_preferences(self) -> TilingPreferences:
        """
        Configure tiling preferences. Return a dictionary with the
        following keys:

        - "depth": number of frames/frame parts to stack on top of each other
        - "total_size": total size of a tile in bytes

        .. versionadded:: 0.6.0
        """
        return {
            "depth": UDF.TILE_DEPTH_DEFAULT,
            "total_size": UDF.TILE_SIZE_MAX,
        }

    def get_backends(self) -> BackendSpec:
        '''
        Signal which computation back-ends the UDF can use.

        Allowed are values in :attr:`libertem.udf.base.UDF.BACKEND_ALL`.
        :code:`(UDF.BACKEND_NUMPY, )` is returned by default for CPU-based computation.

        .. versionadded:: 0.6.0

        .. versionchanged:: 0.11.0
            Extended from just NumPy vs. CuPy to include more generic specification of backends,
            including sparse arrays. See :ref:`sparse` for details! The backend specifications are
            backwards-compatible with the previously supported values.

        Returns
        -------

        backends
            A single value or iterable containing values from the supported backends
            in :attr:`libertem.udf.base.UDF.BACKEND_ALL`
        '''
        return (UDF.BACKEND_NUMPY, )

    def forbuf(self, arr, target):
        '''
        Convert array to backend that is compatible with result buffers and
        reshape.

        This function should be wrapped around assignment to result buffers if
        the array :code:`arr` might have an incompatible array backend and/or
        have a flattened sig dimension. It converts any of the supported input
        backends into the array backend of the target buffer. If the argument is
        already matching the buffer or is not recognized, it is a no-op.

        In particular, the result of array operations on a sparse input tile
        can't always be merged into dense result buffers directly since arrays
        from the :mod:`sparse` package are not converted to NumPy arrays
        automatically.

        .. versionadded:: 0.11.0

        Parameters
        ----------

        arr : array-like
            Any array-like object that is supported by the `sparseconverter`
            package or allows the desired operation with the result buffer at
            hand.
        target : array-like
            The LiberTEM result buffer that the desired operation targets.

        Examples
        --------

        >>> def process_tile(self, tile):
        ...     res = ...  # Processing result from tile
        ...     self.results.intensity[:] += self.forbuf(
        ...         res,
        ...         self.results.intensity
        ...     )

        See :ref:`sparse` for a complete example how to use this function in the
        contaxt of a UDF!
        '''
        res = make_like(arr, target, strict=False)
        return res

    def cleanup(self) -> None:  # FIXME: name? implement cleanup as context manager somehow?
        pass

    @staticmethod
    def with_mask(data: np.ndarray, mask: Union[np.ndarray, bool]) -> ArrayWithMask:
        """
        Add a mask to indicate the valid parts in a result. The mask
        should be an array of bools, with a shape matching
        the `data` array. The mask should have `True` in positions
        that are valid, and `False` otherwise.

        You can pass `mask=True` or `mask=False` as shortcuts for
        all-valid or all-invalid, and also pass in masks that can
        be broadcast to `data`.

        This should be used in :meth:`get_results`.

        See the :ref:`udf valid data masking` section in the
        documentation for details and more examples.

        Example
        -------

        >>> class SomeUDF(UDF):
        ...     # ... other methods omitted for brevity ...
        ...
        ...     def get_results(self):
        ...         return {
        ...             'some_result': self.with_mask(
        ...                 data=self.results.some_result,
        ...                 mask=False,
        ...             ),
        ...         }
        ...
        """
        return ArrayWithMask(data, mask=mask)

    def buffer(
        self,
        kind: BufferKind,
        extra_shape: tuple[int, ...] = (),
        dtype: "nt.DTypeLike" = "float32",
        where: BufferLocation = None,
        use: Optional[BufferUse] = None,
    ) -> BufferWrapper:
        '''
        Use this method to create :class:`~ libertem.common.buffers.BufferWrapper` objects
        in :meth:`get_result_buffers`.

        Parameters
        ----------
        kind : "nav", "sig" or "single"
            The abstract shape of the buffer, corresponding either to the navigation
            or the signal dimensions of the dataset, or a single value.

        extra_shape : optional, tuple of int or a Shape object
            You can specify additional dimensions for your data. For example, if
            you want to store 2D coords, you would specify :code:`(2,)` here.
            If this is specified as a Shape object, it is converted to a tuple first.

        dtype : string or numpy dtype
            The dtype of this buffer

        where : string or None
            :code:`None` means NumPy array, specify :code:`'device'` to use a device buffer
            (for example on a CUDA device)

            .. versionadded:: 0.6.0

        use : "private", "result_only" or None
            If you specify :code:`"private"` here, the result will only be made available
            to internal functions, like :meth:`process_frame`, :meth:`merge` or
            :meth:`get_results`. It will not be available to the user of the UDF, which means
            you can use this to hide implementation details that are likely to change later.

            Specify :code:`"result_only"` here if the buffer is only used in :meth:`get_results`,
            this means we don't have to allocate and return it on the workers without actually
            needing it.

            :code:`None` means the buffer is used both as a final and intermediate result.

            .. versionadded:: 0.7.0
        '''
        if use is not None and use.lower() == "result_only":
            return PlaceholderBufferWrapper(kind, extra_shape, dtype, use=use)
        return BufferWrapper(kind, extra_shape, dtype, where, use=use)

    @classmethod
    def aux_data(cls, data, kind, extra_shape=(), dtype="float32"):
        """
        Use this method to create auxiliary data. Auxiliary data should
        have a shape like `(dataset.shape.nav, extra_shape)` and on access,
        an appropriate view will be created. For example, if you access
        aux data in `process_frame`, you will get the auxiliary data for
        the current frame you are processing.

        Example
        -------

        We create a UDF to demonstrate the behavior:

        >>> class MyUDF(UDF):
        ...     def get_result_buffers(self):
        ...         # Result buffer for debug output
        ...         return {'aux_dump': self.buffer(kind='nav', dtype='object')}
        ...
        ...     def process_frame(self, frame):
        ...         # Extract value of aux data for demonstration
        ...         self.results.aux_dump[:] = str(self.params.aux_data[:])
        ...
        >>> # for each frame, provide three values from a sequential series:
        >>> aux1 = MyUDF.aux_data(
        ...     data=np.arange(np.prod(dataset.shape.nav) * 3, dtype=np.float32),
        ...     kind="nav", extra_shape=(3,), dtype="float32"
        ... )
        >>> udf = MyUDF(aux_data=aux1)
        >>> res = ctx.run_udf(dataset=dataset, udf=udf)

        process_frame for frame (0, 7) received a view of aux_data with values [21., 22., 23.]:

        >>> res['aux_dump'].data[0, 7]
        '[21. 22. 23.]'
        """
        buf = AuxBufferWrapper(kind, extra_shape, dtype)
        buf.set_buffer(data)
        return buf


class NoOpUDF(UDF):
    '''
    A UDF that does nothing and returns nothing.

    This is useful for testing.

    Parameters
    ----------
    preferred_input_dtype : numpy.dtype
        Perform dtype conversion. By default, this is :attr:`UDF.USE_NATIVE_DTYPE`.
    '''
    def __init__(self, preferred_input_dtype=UDF.USE_NATIVE_DTYPE) -> None:
        super().__init__(preferred_input_dtype=preferred_input_dtype)

    def process_tile(self, tile):
        '''
        Do nothing.
        '''
        pass

    def get_result_buffers(self):
        '''
        No result buffers.
        '''
        return {}

    def get_preferred_input_dtype(self):
        '''
        Return the value passed in the constructor.
        '''
        return self.params.preferred_input_dtype


def check_cast(fromvar, tovar):
    if not np.can_cast(fromvar.dtype, tovar.dtype, casting='safe'):
        # FIXME exception or warning?
        raise TypeError(f"Unsafe automatic casting from {fromvar.dtype} to {tovar.dtype}")


class UDFParams:
    def __init__(
        self,
        kwargs: list[dict],
        roi: Optional[np.ndarray],
        corrections: Optional[CorrectionSet],
        tiling_scheme: TilingScheme,
    ):
        """
        Container class for UDF parameters for multiple UDFs

        Parameters
        ----------
        kwargs : List[dict]
            List of kwargs which can be used to re-create the UDF instances
            that were passed to `run_for_dataset`
        roi
            Boolean array to select parts of the navigation space
        corrections
            Corrections to apply
        """
        self._kwargs = kwargs
        self._roi = roi
        self._corrections = corrections
        self._tiling_scheme = tiling_scheme

    @classmethod
    def from_udfs(
        cls,
        udfs: Iterable[UDF],
        roi: Optional[np.ndarray],
        corrections: Optional[CorrectionSet],
        tiling_scheme: TilingScheme,
    ):
        kwargs = [udf._kwargs for udf in udfs]
        return cls(kwargs, roi, corrections, tiling_scheme)

    def patch(self, patch):
        for p, kwargs in zip(patch, self._kwargs):
            kwargs.update(p)

    @property
    def roi(self):
        return self._roi

    @property
    def corrections(self):
        return self._corrections

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def tiling_scheme(self):
        return self._tiling_scheme


class Task:
    """
    A computation on a partition. Inherit from this class and implement ``__call__``
    for your specific computation.

    .. versionchanged:: 0.4.0
        Moved from libertem.job.base to libertem.udf.base as part of Job API deprecation
    """

    def __init__(self, partition: Partition, idx: int, span_context: "SpanContext"):
        self.partition = partition
        self.idx = idx
        self._span_context = span_context
        self._progress = False

    def get_partition(self):
        return self.partition

    def get_locations(self):
        return self.partition.get_locations()

    def get_resources(self) -> ResourceDef:
        '''
        Specify the resources that a Task will use.

        The resources are designed to work with resource tags in Dask clusters:
        See https://distributed.dask.org/en/latest/resources.html

        The resources allow scheduling of CPU-only compute, CUDA-only compute,
        (CPU or CUDA) hybrid compute, and service tasks in such a way that all
        resources are used without oversubscription. Furthermore, they
        distinguish if the given resource can be accessed with a transparent
        NumPy ndarray interface -- namely if CuPy is installed to access CUDA
        resources.

        Each CPU worker gets one CPU, one compute and one ndarray resource
        assigned. Each CUDA worker gets one CUDA and one compute resource
        assigned. If CuPy is installed, it additionally gets an ndarray resource
        assigned. A service worker doesn't get any resources assigned.

        A CPU-only task consumes one CPU, one ndarray and one compute resource,
        i.e. it will be scheduled only on CPU workers. A CUDA-only task consumes
        one CUDA, one compute and possibly an ndarray resource, i.e. it will
        only be scheduled on CUDA workers. A hybrid task that can run on both
        CPU or CUDA using a transparent ndarray interface only consumes a
        compute and an ndarray resource, i.e. it can be scheduled on CPU workers
        or CUDA workers with CuPy. A service task doesn't request any resources
        and can therefore run on CPU, CUDA and service workers.

        Which device a hybrid task uses is only decided at runtime by the
        environment variables that are set on each worker process.

        That way, CPU-only and CUDA-only tasks can run in parallel without
        oversubscription, and hybrid tasks use whatever compute workers are free
        at the time. Furthermore, it allows using CUDA without installing CuPy.
        See https://github.com/LiberTEM/LiberTEM/pull/760 for detailed
        discussion.

        Compute tasks will not run on service workers, so that these can serve
        shorter service tasks with lower latency.

        At the moment, workers only get a single "item" of each resource
        assigned since we run one process per CPU. Requesting more of a resource
        than any of the workers has causes a :class:`RuntimeError`, including
        requesting a resource that is not available at all.

        For that reason one has to make sure that the workers are set up with
        the correct resources and matching environment.
        :meth:`libertem.utils.devices.detect`,
        :meth:`libertem.executor.dask.cluster_spec` and
        :meth:`libertem.executor.dask.DaskJobExecutor.make_local` can be used to
        set up a local cluster correctly.

        .. versionadded:: 0.6.0
        '''
        # default: run only on CPU and do computation
        # No CUDA support for the deprecated Job interface
        return {'CPU': 1, 'compute': 1, 'ndarray': 1}

    def get_tracing_span_context(self) -> "SpanContext":
        return self._span_context

    def __call__(self, params: UDFParams, env: Environment):
        raise NotImplementedError()

    def report_progress(self):
        self._progress = True


def _get_canonical_backends(backends: Optional[BackendSpec]) -> Sequence[ArrayBackend]:
    """
    Convert from either an iterable of backends or a simple string form into a
    canonical form of Sequence[str]
    """
    if backends is None:
        return ()
    if isinstance(backends, str):
        backends = (backends,)
    return tuple(backends)


def get_resources_for_backends(
    udf_backends: list[BackendSpec],
    user_backends: Optional[BackendSpec]
) -> ResourceDef:
    """
    Find the resource definition that is appropriate for the backends specified
    by the UDFs and the user. This is the intersection between the backend specified
    in each UDF, and the backend that was sepected by the user.

    Parameters
    ----------
    udf_backends
        The backends specified by each UDF that we want to run

    user_backends
        The backends specified by the user

    See :meth:`Task.get_resources` for details.
    """
    udf_backends_canonical = [
        _get_canonical_backends(bs)
        for bs in udf_backends
    ]
    user_backends_canonical = _get_canonical_backends(user_backends)

    needs_cuda = 0
    needs_cpu = 0
    needs_ndarray = 0

    # Limit to externally specified backends
    for backend_set in udf_backends_canonical:
        if user_backends_canonical:
            backends = set(user_backends_canonical).intersection(backend_set)
        else:
            backends = set(backend_set)
        needs_cuda += backends.isdisjoint(CPU_BACKENDS)
        needs_cpu += backends.isdisjoint(CUDA_BACKENDS)
        needs_ndarray += CUDA not in backends
    if needs_cuda and needs_cpu:
        raise ValueError(
            "There is no common supported UDF backend (have: %r, limited to %r)"
            % (udf_backends, user_backends_canonical)
        )
    result: ResourceDef = {'compute': 1}
    if needs_cpu:
        result['CPU'] = 1
    if needs_cuda:
        result['CUDA'] = 1
    if needs_ndarray:
        result['ndarray'] = 1
    return result


class UDFTask(Task):
    def __init__(
        self,
        partition: Partition,
        idx: int,
        udf_classes: list[type[UDF]],
        udf_backends: list[BackendSpec],
        user_backends: Optional[BackendSpec],
        runner_cls: type['UDFPartRunner'],
        span_context: "SpanContext",
        task_frames: int,
    ):
        """
        A computation for a single partition. The parameters that stay the same
        for the whole dataset are excluded here and supplied by the executor in
        __call__.

        Parameters
        ----------
        partition : Partition
            The partition to work on
        idx : int
            the index of the task, used to identify results (?)
        udf_classes : List[Type[UDF]]
            The UDFs to run
        backends : List[str]
            The specified backends we want to run on
        task_frames : int
            The number of frames this task must process (ROI included)
        """
        super().__init__(partition=partition, idx=idx, span_context=span_context)
        self._udf_classes = udf_classes
        self._udf_backends = udf_backends
        self._user_backends = user_backends
        self._runner_cls = runner_cls
        self._task_frames = task_frames

    def __call__(self, params: UDFParams, env: Environment) -> tuple[UDFData, ...]:
        with self._propagate_tracing(), tracer.start_as_current_span("UDFTask.__call__"):
            udfs = [
                cls.new_for_partition(kwargs, self.partition, params.roi)
                for cls, kwargs in zip(self._udf_classes, params.kwargs)
            ]
            return self._runner_cls(udfs, progress=self._progress).run_for_partition(
                self.partition, params, env, self._user_backends,
            )

    @contextmanager
    def _propagate_tracing(self):
        """
        Initialize the current tracing context from the :code:`SpanContext`
        given in `self._span_context`, if we don't already have a running
        tracing context.
        """
        if opentelemetry_context.get_current():
            yield
        else:
            with attach_to_parent(self._span_context):
                yield

    def get_runner_cls(self) -> type["UDFPartRunner"]:
        return self._runner_cls

    def get_udf_classes(self) -> list[type["UDF"]]:
        return self._udf_classes

    def get_resources(self) -> ResourceDef:
        """
        Intersection of resources of all UDFs, throws if empty.

        See docstring of super class for details.
        """
        return get_resources_for_backends(self._udf_backends, user_backends=self._user_backends)

    def __repr__(self):
        return f"<UDFTask {self._udf_classes!r}>"

    @property
    def task_frames(self) -> int:
        """
        The number of frames to be processed in this task
        (accounting for ROI)
        """
        return self._task_frames


class UDFPartRunner:
    def __init__(self, udfs: list[UDF], debug: bool = False, progress: bool = False):
        self._udfs = udfs
        self._debug = debug
        self._progress = progress

    def run_for_partition(
        self,
        partition: Partition,
        params: UDFParams,
        env: Environment,
        backend_choice: BackendSpec
    ) -> tuple[UDFData, ...]:
        roi = params.roi
        corrections = params.corrections
        if backend_choice is None:
            backend_choice = BACKENDS
        backend_choice = frozenset(_get_canonical_backends(backend_choice))
        with env.enter():
            try:
                previous_id = None
                device_class = get_device_class()
                if device_class == 'cpu':
                    available_backends = backend_choice.intersection(CPU_BACKENDS)
                elif device_class == 'cuda':
                    if has_cupy():
                        available_backends = backend_choice.intersection(BACKENDS)
                    else:
                        available_backends = backend_choice.intersection(
                            CPU_BACKENDS.union((CUDA, ))
                        )
                else:
                    raise RuntimeError(f"Unknown device class {device_class}.")
                # This will raise an exception if unavailable backends would be used
                ds_backend, execution_plan = _execution_plan(
                    udfs=self._udfs,
                    ds=partition.meta,
                    device_class=device_class,
                    available_backends=available_backends,
                )
                if CUPY_BACKENDS.intersection(execution_plan.keys()):
                    # Avoid importing if not used
                    import cupy
                    device = get_use_cuda()
                    previous_id = cupy.cuda.Device().id
                    cupy.cuda.Device(device).use()
                dtype = self._init_udfs(
                    execution_plan, partition, roi, corrections, device_class, env,
                    params.tiling_scheme,
                )
                partition.set_corrections(corrections)
                if env.worker_context is not None:
                    partition.set_worker_context(env.worker_context)
                self._run_udfs(
                    ds_backend, execution_plan, partition, params.tiling_scheme, roi, dtype
                )
                self._wrapup_udfs(partition)
            finally:
                if previous_id is not None:
                    cupy.cuda.Device(previous_id).use()
            # Make sure results are in the same order as the UDFs
            return tuple(udf.results for udf in self._udfs)

    def _run_udfs(
        self,
        ds_backend: ArrayBackend,
        execution_plan: ExecutionPlan,
        partition: Partition,
        tiling_scheme: TilingScheme,
        roi: Optional[np.ndarray],
        dtype,
    ) -> None:
        # FIXME pass information on target location (numpy or cupy)
        # to dataset so that is can already move it there.
        # In the future, it might even decode data on the device instead of CPU
        tiles = partition.get_tiles(
            tiling_scheme=tiling_scheme,
            roi=roi, dest_dtype=dtype,
            array_backend=ds_backend,
        )

        # type explicitly to help mypy
        partition_progress: Union[PartitionProgressTracker, PartitionTrackerNoOp]
        if self._progress:
            partition_progress = PartitionProgressTracker(partition)
        else:
            partition_progress = PartitionTrackerNoOp()
        partition_progress.signal_start()

        # Save udf process_ method to avoid repeated isinstance() checks
        # in the future could be computed along with the execution_plan
        # or better still computed on the main node to ensure consistency
        # with the tiling scheme
        udf_methods = {
            backend: tuple(udf.get_method() for udf in udfs)
            for backend, udfs in execution_plan.items()
        }

        for backend, udfs in execution_plan.items():
            for udf in udfs:
                udf.meta.set_valid_nav_mask(None)

        try:
            for tile in tiles:
                converter = TileConverter(tile)
                for backend, udfs in execution_plan.items():
                    methods = udf_methods[backend]
                    device_tile = converter.get(backend)
                    self._run_tile(
                        zip(udfs, methods), partition, tile, backend, device_tile, roi=roi
                    )
                partition_progress.signal_tile_complete(tile)
        except AttributeError as e:
            removed = {
                'tile_slice': 'self.meta.slice',
                'scheme_idx': 'self.meta.tiling_scheme_idx',
            }
            for r, repl in removed.items():
                if r in e.args[0]:
                    raise AttributeError(
                        f'Attribute {r} for input tiles was removed. Please use {repl} instead.'
                    )
            raise

        # We could signal partition completion here, but this is already
        # handled on the main node via ProgressManager.finalize_task(task)
        # as it is the fallback method for executors that don't support comms
        # partition_progress.signal_complete()

    def _init_udfs(
        self,
        execution_plan: ExecutionPlan,
        partition: Partition,
        roi: Optional[np.ndarray],
        corrections: CorrectionSet,
        device_class,
        env: Environment,
        tiling_scheme: TilingScheme,
    ) -> "nt.DTypeLike":
        dtype = _get_dtype(self._udfs, partition.dtype, corrections, execution_plan.keys())
        for backend, udfs in execution_plan.items():
            pslice_for_roi = partition.slice.adjust_for_roi(roi)
            meta = UDFMeta(
                partition_slice=pslice_for_roi,
                dataset_shape=partition.meta.shape,
                roi=roi,
                dataset_dtype=partition.dtype,
                input_dtype=dtype,
                tiling_scheme=tiling_scheme,
                corrections=corrections,
                device_class=device_class,
                threads_per_worker=env.threads_per_worker,
                array_backend=backend,
            )
            for udf in udfs:
                udf.set_backend(backend)
                udf.set_meta(meta)
                # validate that get_method returns a valid method type
                # the presence of the method was checked on the main node
                if udf.get_method() not in tuple(UDFMethod):  # pragma: no cover
                    # Should never be reached
                    raise UDFException('UDF.get_method() returned unrecognized')
                udf.set_slice(pslice_for_roi)
                udf.init_result_buffers()
                udf.allocate_for_part(partition, roi)
                udf.init_task_data()
                if isinstance(udf, UDFPreprocessMixin):
                    udf.clear_views()
                    udf.preprocess()

        return dtype

    def _run_tile(
        self,
        udfs_and_methods: Iterable[
            tuple[
                UDF,
                UDFMethod,
            ]
        ],
        partition: Partition,
        tile: DataTile,
        array_backend: ArrayBackend,
        device_tile: Any,
        roi: Optional[np.ndarray],
    ) -> None:
        for udf, udf_method in udfs_and_methods:
            if udf_method == UDFMethod.TILE:
                udf.set_contiguous_views_for_tile(partition, tile)
                udf.set_slice(tile.tile_slice)
                udf.set_tile_idx(tile.scheme_idx)
                udf_tile = typing.cast(UDFTileMixin, udf)
                udf_tile.process_tile(device_tile)
            elif udf_method == UDFMethod.FRAME:
                tile_slice = tile.tile_slice
                udf_frame = typing.cast(UDFFrameMixin, udf)
                for frame_idx, frame in enumerate(device_tile):
                    frame_slice = Slice(
                        origin=(tile_slice.origin[0] + frame_idx,) + tile_slice.origin[1:],
                        shape=Shape((1,) + tuple(tile_slice.shape)[1:],
                                    sig_dims=tile_slice.shape.sig.dims),
                    )
                    # Internal checks for dataset consistency
                    assert frame_slice.shape[0] == 1
                    # assert .... or True to allow disabling this check
                    # in a production run for performance
                    if array_backend in ND_BACKENDS:
                        # The sig dimension is squeezed out
                        assert check_shape(frame, frame_slice.shape[1:]) or True
                    elif array_backend in D2_BACKENDS:
                        # The matrix backend of the 2D arrays adds a sig dimension of 1
                        assert check_shape(frame, frame_slice.shape) or True
                    udf.set_slice(frame_slice)
                    udf.set_views_for_frame(partition, tile, frame_idx)
                    udf_frame.process_frame(frame)
            elif udf_method == UDFMethod.PARTITION:
                # Internal checks for dataset consistency
                assert partition.slice.adjust_for_roi(roi) == tile.tile_slice
                udf.set_views_for_tile(partition, tile)
                udf.set_slice(tile.tile_slice)
                udf_partition = typing.cast(UDFPartitionMixin, udf)
                udf_partition.process_partition(device_tile)
            else:  # pragma: no cover
                # Should never be reached!
                raise UDFException('UDF.get_method() return unrecognized '
                                   'method during _run_udfs')

    def _wrapup_udfs(
        self,
        partition: Partition
    ) -> None:
        for udf in self._udfs:
            udf.flush(self._debug)
            if isinstance(udf, UDFPostprocessMixin):
                udf.clear_views()
                udf.postprocess()

            udf.cleanup()
            udf.clear_views()
            udf.export_results()

        if self._debug:
            try:
                cloudpickle.loads(cloudpickle.dumps(partition))
            except TypeError:
                raise TypeError("could not pickle partition")
            try:
                cloudpickle.loads(cloudpickle.dumps(
                    [u.results for u in self._udfs]
                ))
            except TypeError:
                raise TypeError("could not pickle results")


class UDFRunner:
    @staticmethod
    def _apply_part_result(
        udfs: Iterable[UDF],
        damage: BufferWrapper,
        part_results: tuple[UDFData, ...],
        task: TaskProtocol,
    ):
        raw_damage = damage.raw_data
        assert raw_damage is not None
        for results, udf in zip(part_results, udfs):
            udf.meta.set_valid_nav_mask(raw_damage)
            udf.set_views_for_partition(task.get_partition())
            udf.merge(
                dest=udf.results.get_proxy(),
                src=results.get_proxy()
            )
        v = damage.get_view_for_partition(task.get_partition())
        v[:] = True

    @staticmethod
    def _make_udf_result(udfs: Iterable[UDF], damage: BufferWrapper) -> "UDFResults":
        raw_damage = damage.raw_data
        assert raw_damage is not None
        for udf in udfs:
            udf.meta.set_valid_nav_mask(raw_damage)
            udf.clear_views()
        return UDFResults(
            buffers=tuple(
                udf._do_get_results()
                for udf in udfs
            ),
            damage=damage
        )

    def __init__(self, udfs: list[UDF], debug: bool = False,
                 progress_reporter: Optional['ProgressReporter'] = None):
        self._udfs = udfs
        self._debug = debug
        self._pool = TracedThreadPoolExecutor(tracer, max_workers=4)
        self._progress_reporter = progress_reporter

    @classmethod
    def inspect_udf(
        cls,
        udf: UDF,
        dataset: DataSet,
        roi: Optional[np.ndarray] = None,
    ) -> dict[str, BufferWrapper]:
        """
        Return result buffer declarations for a given UDF/DataSet/roi combination
        """
        ds_backend, execution_plan = _execution_plan(
            udfs=(udf, ),
            ds=dataset,
            device_class=None,
            # For the sake of dtype determination we'll assume all backends are available.
            # On the workers some might be missing. The cupyx.scipy.sparse backends might
            # trigger an up-conversion to float that may or may not happen in reality.
            # By being inclusive with backends here we make sure that the result buffers
            # can hold whatever the workers may be sending.
            available_backends=BACKENDS,
        )
        dtype = _get_dtype(
            [udf],
            dataset.dtype,
            corrections=None,
            array_backends=execution_plan.keys(),
        )
        meta = UDFMeta(
            partition_slice=None,
            dataset_shape=dataset.shape,
            roi=None,
            dataset_dtype=dataset.dtype,
            input_dtype=dtype,
            corrections=None,
        )

        udf = udf.copy()
        udf.set_meta(meta)
        buffers = udf.get_result_buffers()
        for buf in buffers.values():
            buf.set_shape_ds(dataset.shape, roi)
        return buffers

    @classmethod
    def dry_run(cls, udfs, dataset, roi=None):
        """
        Return result buffers for a given UDF/DataSet/roi combination
        exactly as running the UDFs would, just skipping execution and
        merging of the processing tasks.

        This can be used to create an empty result to initialize live plots
        before running an UDF.
        """
        with tracer.start_as_current_span("UDFRunner.dry_run"):
            runner = cls(udfs)
            executor = InlineJobExecutor()
            res = runner.run_for_dataset(
                dataset=dataset,
                executor=executor,
                roi=roi,
                dry=True
            )
            return res

    def _debug_task_pickling(self, tasks: list[UDFTask]) -> None:
        if self._debug:
            cloudpickle.loads(cloudpickle.dumps(tasks))

    def _check_preconditions(self, dataset: DataSet, roi: Optional[np.ndarray]) -> None:
        if roi is not None and prod(roi.shape) != prod(dataset.shape.nav):
            raise ValueError(
                "roi: incompatible shapes: {} (roi) vs {} (dataset)".format(
                    roi.shape, dataset.shape.nav
                )
            )

    def _prepare_run_for_dataset(
        self,
        dataset: DataSet,
        executor: JobExecutor,
        roi: Optional[np.ndarray],
        corrections: Optional[CorrectionSet],
        backends: Optional[BackendSpec],
        dry: bool,
    ) -> tuple[list[UDFTask], UDFParams]:
        self._check_preconditions(dataset, roi)
        if backends is None:
            backends = BACKENDS
        backends = _get_canonical_backends(backends)
        ds_backend, execution_plan = _execution_plan(
            udfs=self._udfs,
            ds=dataset,
            # To include all possible backends without preference for device class
            device_class=None,
            # For the sake of dtype determination we'll assume all backends are available.
            # On the workers some might be missing. The cupyx.scipy.sparse backends might
            # trigger an up-conversion to float that may or may not happen in reality.
            # By being inclusive with backends here we make sure that the result buffers
            # can hold whatever the workers may be sending.
            available_backends=backends,
        )
        dtype = _get_dtype(self._udfs, dataset.dtype, corrections, execution_plan.keys())
        meta = UDFMeta(
            partition_slice=None,
            dataset_shape=dataset.shape,
            roi=roi,
            dataset_dtype=dataset.dtype,
            input_dtype=dtype,
            corrections=corrections,
            # We don't specify the dataset input backend here since
            # that will be determined dynamically on the workers.
        )
        for udf in self._udfs:
            udf.set_meta(meta)
            # ideally the UDF would know the set of workers it might run on
            # here, e.g. backend information, to choose the best method,
            # as-is the backend will always be numpy by default
            udf_method = udf.get_method()
            if udf_method not in tuple(UDFMethod):
                raise UDFException('UDF.get_method() returned unrecognized value')
            elif udf_method == UDFMethod.TILE and not isinstance(udf, UDFTileMixin):
                raise UDFException(f'UDF declared method {udf_method.value} but '
                                   'does not implement process_tile.')
            elif udf_method == UDFMethod.FRAME and not isinstance(udf, UDFFrameMixin):
                raise UDFException(f'UDF declared method {udf_method.value} but '
                                   'does not implement process_frame.')
            elif udf_method == UDFMethod.PARTITION and not isinstance(udf, UDFPartitionMixin):
                raise UDFException(f'UDF declared method {udf_method.value} but '
                                   'does not implement process_partition.')
            udf.init_result_buffers(executor=executor)
            udf.allocate_for_full(dataset, roi)

            if isinstance(udf, UDFPreprocessMixin):
                udf.set_views_for_dataset(dataset)
                udf.preprocess()
        with tracer.start_as_current_span('tileshape negotiation'):
            neg = Negotiator()
            # FIXME take compute backend into consideration as well
            # Other boundary conditions when moving input data to device
            # FIXME: approximate partition shape here
            partition = next(dataset.get_partitions())
            tiling_scheme = neg.get_scheme(
                udfs=self._udfs,
                approx_partition_shape=partition.shape,
                dataset=dataset,
                read_dtype=meta.input_dtype,
                roi=roi,
                corrections=corrections,
            )
        params = UDFParams.from_udfs(
            udfs=self._udfs,
            roi=roi,
            corrections=corrections,
            tiling_scheme=tiling_scheme,
        )
        if dry:
            tasks = []
        else:
            tasks = list(self._make_udf_tasks(dataset, roi, backends))
        return (tasks, params)

    def run_for_dataset(
        self,
        dataset: DataSet,
        executor: JobExecutor,
        roi: Optional[np.ndarray] = None,
        progress: bool = False,
        corrections: Optional[CorrectionSet] = None,
        backends: Optional[BackendSpec] = None,
        dry: bool = False
    ) -> "UDFResults":
        with tracer.start_as_current_span("UDFRunner.run_for_dataset"):
            for res in self.run_for_dataset_sync(
                dataset=dataset,
                executor=executor.ensure_sync(),
                roi=roi,
                progress=progress,
                corrections=corrections,
                backends=backends,
                dry=dry,
                iterate=False
            ):
                pass
            return res  # type: ignore

    def results_for_dataset_sync(
        self,
        dataset: DataSet,
        executor: JobExecutor,
        roi: Optional[np.ndarray] = None,
        progress: bool = False,
        corrections: Optional[CorrectionSet] = None,
        backends: Optional[BackendSpec] = None,
        dry: bool = False,
    ) -> tuple[
        Iterable[tuple[tuple[UDFData, ...], TaskProtocol]],
        Any,
        UDFParams
    ]:
        with tracer.start_as_current_span("_prepare_run_for_dataset"):
            tasks, params = self._prepare_run_for_dataset(
                dataset, executor, roi, corrections, backends, dry
            )

        with tracer.start_as_current_span("before inner work"):
            cancel_id = str(uuid.uuid4())
            self._debug_task_pickling(tasks)

            executor = executor.ensure_sync()
            if dry:
                task_comm_handler: TaskCommHandler = NoopCommHandler()
            else:
                task_comm_handler = dataset.get_task_comm_handler()

        def _inner():
            pman = None
            try:
                if progress and tasks:
                    pman = ProgressManager(tasks, cancel_id, reporter=self._progress_reporter)
                    pman.connect(task_comm_handler)
                    for task in tasks:
                        task.report_progress()
                with executor.scatter(params) as params_handle:
                    # XXX yuck, refactor this?
                    yield params_handle

                    if tasks:
                        for res in executor.run_tasks(
                            tasks,
                            params_handle,
                            cancel_id,
                            task_comm_handler,
                        ):
                            if progress:
                                pman.finalize_task(res[1])
                            yield res
            finally:
                if progress and tasks and pman is not None:
                    pman.close()

        it = _inner()
        params_handle = next(it)
        return it, params_handle, params

    def run_for_dataset_sync(
        self,
        dataset: DataSet,
        executor: JobExecutor,
        roi: Optional[np.ndarray] = None,
        progress: bool = False,
        corrections: Optional[CorrectionSet] = None,
        backends: Optional[BackendSpec] = None,
        dry: bool = False,
        iterate: bool = True
    ) -> ResultsForDataSet:
        executor = executor.ensure_sync()
        result_iter, params_handle, params = self.results_for_dataset_sync(
            dataset=dataset,
            executor=executor,
            roi=roi,
            progress=progress,
            corrections=corrections,
            backends=backends,
            dry=dry,
        )
        damage = BufferWrapper(kind='nav', dtype=bool)
        damage.set_shape_ds(dataset.shape, roi)
        damage.allocate()

        def _inner():
            num_results = 0
            try:
                for part_results, task in result_iter:
                    num_results += 1
                    with tracer.start_as_current_span("_apply_part_result -> UDF.merge"):
                        self._apply_part_result(
                            udfs=self._udfs,
                            damage=damage,
                            part_results=part_results,
                            task=task
                        )
                    if iterate:
                        yield self._make_udf_result(
                            udfs=self._udfs,
                            damage=damage
                        )
                if num_results == 0 or not iterate:
                    yield self._make_udf_result(
                        udfs=self._udfs,
                        damage=damage
                    )
            except JobCancelledError:
                raise UDFRunCancelled(f"UDF run cancelled after {num_results} partitions")

        gen = _inner()
        return ResultsForDataSet(
            gen=gen,
            params_handle=params_handle,
            params=params,
            executor=executor,
            udfs=self._udfs,
        )

    async def run_for_dataset_async(
        self,
        dataset: DataSet,
        executor: JobExecutor,
        cancel_id,
        roi: Optional[np.ndarray] = None,
        corrections: Optional[CorrectionSet] = None,
        backends: Optional[BackendSpec] = None,
        progress: bool = False,
        dry: bool = False,
        iterate: bool = True,
    ) -> AsyncGenerator["UDFResults", None]:
        with tracer.start_as_current_span("UDFRunner.run_for_dataset_async"):
            gen = self.run_for_dataset_sync(
                dataset=dataset,
                executor=executor.ensure_sync(),
                roi=roi,
                progress=progress,
                corrections=corrections,
                backends=backends,
                dry=dry,
                iterate=iterate,
            )

            async for res in async_generator_eager(gen, pool=self._pool):
                yield res

    def _roi_for_partition(self, roi: np.ndarray, partition: Partition) -> np.ndarray:
        return roi_for_partition(roi, partition)

    def _make_udf_tasks(
        self,
        dataset: DataSet,
        roi: Optional[np.ndarray],
        backends: Optional[BackendSpec]
    ) -> Generator[UDFTask, None, None]:
        udf_backends = [
            udf.get_backends()
            for udf in self._udfs
        ]

        span = trace.get_current_span()
        span_context = span.get_span_context()
        for idx, partition in enumerate(dataset.get_partitions()):
            partition.set_idx(idx)
            num_frames = partition.get_frame_count(roi)
            if not num_frames:
                # roi is empty for this partition or zero-length, ignore
                continue
            udf_classes = [
                udf.__class__
                for udf in self._udfs
            ]
            tasks = UDFTask(
                partition=partition, idx=idx, udf_classes=udf_classes,
                udf_backends=udf_backends,
                user_backends=backends,
                runner_cls=self.get_part_runner_cls(),
                span_context=span_context,
                task_frames=num_frames,
            )
            yield tasks

    @classmethod
    def get_part_runner_cls(cls) -> type['UDFPartRunner']:
        return UDFPartRunner


UDFResultDict = Mapping[str, BufferWrapper]


class UDFResults:
    '''
    Container class to combine UDF results with additional information.

    This class allows to return additional information from UDF execution
    together with UDF result buffers. This is currently used to pass
    "damage" information when running UDFs as an iterator using
    :meth:`libertem.api.Context.run_udf_iter`. "Damage" is
    a map of the nav space that is set to :code:`True`
    for all positions that have already been processed.

    .. versionadded:: 0.7.0

    Parameters
    ----------

    buffers
        Iterable containing the result buffer dictionaries for each of the UDFs being executed

    damage : BufferWrapper
        :class:`libertem.common.buffers.BufferWrapper` of :code:`kind='nav'`, :code:`dtype=bool`.
        It is set to :code:`True` for all positions in nav space that have been processed already.
    '''
    def __init__(self, buffers: Iterable[UDFResultDict], damage: BufferWrapper):
        self.buffers = tuple(buffers)
        self.damage = damage



import dask.array as da
import numpy as np

from ...common.math import prod
from ...common.slice import Slice
from ...common.buffers import BufferWrapper
from .dask_inplace import DaskInplaceWrapper


class DaskBufferWrapper(BufferWrapper):
    @classmethod
    def from_buffer(cls, buffer: BufferWrapper):
        """
        Convert an existing BufferWrapper into a DaskBufferWrapper
        without copying its data, only its parameters
        """
        if buffer.has_data():
            RuntimeWarning('Buffer is already allocated, not copying current contents or view')
        result = cls(kind=buffer.kind,
                   extra_shape=buffer.extra_shape,
                   dtype=buffer.dtype,
                   where=buffer.where,
                   use=buffer.use)
        result._shape = buffer._shape
        result._ds_shape = buffer._ds_shape
        result._ds_partitions = buffer._ds_partitions
        result._roi = buffer._roi
        result._roi_is_zero = buffer._roi_is_zero
        return result

    @property
    def _extra_chunking(self):
        return tuple((s,) for s in self._extra_shape)

    def allocate(self, lib=None):
        """
        Allocate a buffer as a dask array

        nav buffer chunking is made identical to the partition structure
        in the flat nav dimension via the attribute self._ds_partitions
        sig buffers are currently unchunked as there is no mechanism
        to have partitions which cover part of the signal dimension

        extra_shape dimensions are currently unchunked
        """
        # self._ds_partitions set in UDFData.allocate_for_full()
        nav_chunking = (tuple(self._slice_for_partition(p).shape[0]
                              for p in self._ds_partitions),)

        # At this time there is no interest in chunking the sig dimension
        # This could be modified in the future if partitions are more flexible
        sig_shape = self._ds_shape.sig
        sig_chunking = (-1,) * len(sig_shape) + self._extra_chunking

        if self._kind == 'nav':
            flat_nav_shape = self._shape[0]
            if self._shape == (flat_nav_shape,) + self._extra_shape:
                _buf_chunking = nav_chunking + self._extra_chunking
            elif self._shape[0] in nav_chunking:
                # This branch should never be taken if we are only allocating
                # with dask on the main node and not inside UDFTasks
                _buf_chunking = self._shape
            else:
                raise RuntimeError('Unrecognized buffer size relative to ds/chunk globals')
        elif self._kind == 'sig':
            _buf_chunking = sig_chunking
        elif self._kind == 'single':
            _buf_chunking = self._shape
        else:
            raise NotImplementedError('Unrecognized buffer kind')
        self._data = da.zeros(self._shape, dtype=self._dtype, chunks=_buf_chunking)

    @property
    def data(self):
        """
        Get the buffer contents in shape that corresponds to the
        original dataset shape, using a lazy Dask array.

        Copied largely from BufferWrapper with modifications to ensure
        Dask arrays are correctly unpacked into the result array.

        #TODO consider if this needs to be cached to avoid creating
        multiple copies in the task graph ?

        If a ROI is set, embed the result into a new
        array; unset values have NaN value for floating point types,
        False for boolean, 0 for integer types and structs,
        '' for string types and None for objects.
        """
        if isinstance(self._data, DaskInplaceWrapper):
            self._data = self._data.data
        if self._contiguous_cache:
            raise RuntimeError("Cache is not empty, has to be flushed")
        if self._roi is None or self._kind != 'nav':
            return self._data.reshape(self._shape_for_kind(self._kind, self._ds_shape))
        shape = self._shape_for_kind(self._kind, self._ds_shape)
        if shape == self._data.shape:
            # preallocated and already wrapped
            return self._data
        # Integer types and "void" (structs and such)
        if self.dtype.kind in ('i', 'u', 'V'):
            fill = 0
        # Bytes and Unicode strings
        elif self.dtype.kind in ('S', 'U'):
            fill = ''
        else:
            # 'b' (boolean): False
            # 'f', 'c': NaN
            # 'm', 'M' (datetime, timedelta): NaT
            # 'O' (object): None
            fill = None

        flat_chunking = tuple(p.slice.shape[0] for p in self._ds_partitions)
        flat_chunking = (flat_chunking,) + self._extra_chunking
        flat_shape = (prod(self._ds_shape.nav),) + self._extra_shape
        flat_wrapper = da.full(flat_shape, fill, dtype=self._dtype, chunks=flat_chunking)
        flat_wrapper[self._roi, ...] = self._data
        wrapper = flat_wrapper.reshape(self._ds_shape.nav + self._extra_shape)
        return wrapper

    def _get_slice(self, slice: Slice):
        """
        Get a view of the buffer which enable inplace assignment
        to the underlying Dask array, for simple inplace assignments

        This is not robust to doing complex operations on the
        buffer such as changing its shape
        """
        real_slice = slice.get()
        inplace_wrapper = DaskInplaceWrapper(self._data)
        inplace_wrapper.set_slice(real_slice)
        return inplace_wrapper

    def export(self):
        """
        No requirement to move dask arrays between devices
        This functionality is embedded in the .compute() call
        """
        self._data = self._data

    def __repr__(self):
        return (f"<{self.__class__.__name__} kind={self._kind} "
                f"extra_shape={self._extra_shape} backing={self._data}>")

    def result_buffer_type(self):
        return self.new_as_prealloc

    def new_as_prealloc(self, *args, **kwargs):
        buffer = DaskPreallocBufferWrapper(*args, **kwargs)
        buffer.add_partitions(self._ds_partitions)
        return buffer


class DaskPreallocBufferWrapper(DaskBufferWrapper):
    def __init__(self, data, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = data


class DaskResultBufferWrapper(DaskPreallocBufferWrapper):
    """
    A DaskBufferWrapper that is meant to be used for results.

    In the LiberTEM internals, we need to make sure everything is still lazy/delayed.

    Once we leave LiberTEM internals, the external interface is used, where
    both :meth:`data` and :meth:`raw_data` eagerly return numpy array-like
    objects.
    """
    @classmethod
    def from_buffer_wrapper(cls, buffer: BufferWrapper):
        assert buffer.has_data()
        result = cls(
            data=buffer._data,
            kind=buffer.kind,
            extra_shape=buffer.extra_shape,
            dtype=buffer.dtype,
            where=buffer.where,
            use=buffer.use
        )
        result._shape = buffer._shape
        result._ds_shape = buffer._ds_shape
        result._ds_partitions = buffer._ds_partitions
        result._roi = buffer._roi
        result._roi_is_zero = buffer._roi_is_zero
        result._valid_mask = buffer._valid_mask
        return result

    def __array__(self):
        return super().__array__()

    @property
    def raw_data(self):
        """
        Eager version of :meth:`delayed_raw_data`
        """
        return np.array(super().raw_data)

    @property
    def delayed_raw_data(self):
        return super().raw_data

    @property
    def data(self):
        """
        Eager version of :meth:`delayed_data`
        """
        return np.array(super().data)

    @property
    def delayed_data(self):
        return super().data



import functools
from typing import Optional

import numpy as np
import numba
import sparse

from libertem.common import Slice
from libertem.io.corrections.detector import correct, RepairDescriptor


@numba.njit(cache=True)
def disjunct_multiplier(excluded, sig_shape, base_shape=1, target=1):
    '''
    Calculate an integer i close to target which is a multiple of base_shape
    and for which i * n not in "excluded" for any n > 0, i * n < sig_shape.

    Cases with a bad pixel at 0 are handled downstream.

    Parameters
    ----------

    excluded: Sequence
        Forbidden tile boundary positions
    sig_shape: int
        Signal shape of the adjusted dimension
    base_shape: int
        Base shape of the adjusted dimension
    target: int
        Target value for the tile size, find a solution close to this value.

    Returns
    -------

    int
    '''
    approx_multiplier = int(np.round(target / base_shape))
    current_value = base_shape * approx_multiplier
    max_excluded = np.max(excluded)
    excluded_map = np.zeros(max_excluded + 1, dtype=np.uint8)
    excluded_map[excluded] = 1
    if current_value >= target:
        sign = 1
    else:
        sign = -1
    for offset in range(max_excluded // base_shape + 1):
        # plus 0, minus 1, plus 2, ...
        current_value += offset * sign * base_shape
        sign *= -1
        if current_value <= 0:
            continue
        clear = True
        for multiplier in range(1, max_excluded // current_value + 1):
            index = current_value * multiplier
            bad = (
                index >= 0
                and index < sig_shape
                and index <= max_excluded
                and excluded_map[index]
            )
            if bad:
                clear = False
                break
        if clear:
            return current_value
    # target value is max_excluded + 1,
    # i.e. if the modulo is 0 we have to add one more
    multiple = max_excluded // base_shape + 1
    return min(multiple * base_shape, sig_shape)


class CorrectionSet:
    """
    A set of corrections to apply.

    .. versionadded:: 0.6.0

    Parameters
    ----------
    dark : np.ndarray
        An array containing a dark frame to substract from all frames,
        its shape needs to match the signal shape of the dataset.

    gain : np.ndarray
        An array containing a gain map to multiply with each frame,
        its shape needs to match the signal shape of the dataset.

    excluded_pixels : sparse.COO
        A "sparse pydata" COO array containing only entries for pixels
        that should be excluded. The shape needs to match the signal
        shape of the dataset. Can also be anything that is directly
        compatible with the :code:`sparse.COO` constructor, for example a
        "roi-like" NumPy array. A :code:`sparse.COO` array can be
        directly constructed from a coordinate array, using
        :code:`sparse.COO(coords=coords, data=1, shape=ds.shape.sig)`
    allow_empty : bool
        Do not throw an exception if a repair environment for an excluded pixel
        is empty. The pixel is left uncorrected in that case.
    """
    def __init__(
        self,
        dark: Optional[np.ndarray] = None,
        gain: Optional[np.ndarray] = None,
        excluded_pixels: Optional[sparse.COO] = None,
        allow_empty: bool = False
    ):
        self._dark = dark
        self._gain = gain
        if excluded_pixels is not None:
            excluded_pixels = sparse.COO(excluded_pixels, prune=True)
        self._excluded_pixels = excluded_pixels
        self._allow_empty = allow_empty
        if not allow_empty and excluded_pixels is not None:
            # Construct the environment for checking so that an exception is thrown
            # when the CorrectionSet is instantiated and not when workers try to apply it.
            _ = RepairDescriptor(
                sig_shape=excluded_pixels.shape,
                excluded_pixels=excluded_pixels.coords,
                allow_empty=False
            )

    def get_dark_frame(self) -> Optional[np.ndarray]:
        return self._dark

    def get_gain_map(self) -> Optional[np.ndarray]:
        return self._gain

    def get_excluded_pixels(self) -> Optional[sparse.COO]:
        return self._excluded_pixels

    def have_corrections(self) -> bool:
        corrs = [
            self.get_dark_frame(),
            self.get_gain_map(),
            self.get_excluded_pixels(),
        ]
        return any(c is not None for c in corrs)

    def apply(self, data: np.ndarray, tile_slice: Slice) -> None:
        """
        Apply corrections in-place to `data`, cropping the
        correction data to the `tile_slice`.
        """
        dark_frame = self.get_dark_frame()
        gain_map = self.get_gain_map()

        if not self.have_corrections():
            return

        sig_slice = tile_slice.get(sig_only=True)

        if dark_frame is not None:
            dark_frame = dark_frame[sig_slice]
        if gain_map is not None:
            gain_map = gain_map[sig_slice]

        correct(
            buffer=data,
            dark_image=dark_frame,
            gain_map=gain_map,
            repair_descriptor=self.repair_descriptor(tile_slice.discard_nav()),
            inplace=True,
            sig_shape=tuple(tile_slice.shape.sig),
            allow_empty=self._allow_empty
        )

    @functools.lru_cache(maxsize=512)
    def repair_descriptor(self, sig_slice):
        excluded_pixels = self.get_excluded_pixels()
        if excluded_pixels is not None:
            excluded_pixels = excluded_pixels[sig_slice.get(sig_only=True)]
            excluded_pixels = excluded_pixels.coords
        return RepairDescriptor(
            sig_shape=tuple(sig_slice.shape.sig),
            excluded_pixels=excluded_pixels,
            allow_empty=self._allow_empty
        )

    def adjust_tileshape(self, tile_shape, sig_shape, base_shape):
        excluded_pixels = self.get_excluded_pixels()
        if excluded_pixels is None:
            return tile_shape
        if excluded_pixels.nnz == 0:
            return tile_shape
        excluded_list = excluded_pixels.coords
        adjusted_shape = np.array(tile_shape)
        sig_shape = np.array(sig_shape)
        base_shape = np.array(base_shape)

        adjust(
            adjusted_shape_inout=adjusted_shape,
            sig_shape=sig_shape,
            base_shape=base_shape,
            excluded_list=excluded_list
        )
        invalid = np.logical_or(
            adjusted_shape <= 0,
            adjusted_shape > sig_shape
        )
        adjusted_shape[invalid] = sig_shape[invalid]
        return tuple(adjusted_shape)


def adjust(adjusted_shape_inout, sig_shape, base_shape, excluded_list):
    '''
    Adjust the tile shape to avoid collisions with patched pixels.

    Find a tile shape that is a multiple of base_shape in such a way that
    excluded pixels are not touching a tile boundary.

    The proposed tile shape in adjusted_shape_inout is used as a starting value.
    If collisions can't be avoided, use sig_shape as tile shape for that dimension.

    Parameters
    ----------

    adjusted_shape_inout: np.ndarray, size n_dim
        Shape to adjust, modified in place
    sig_shape: sequence, size n_dim
        Signal shape
    base_shape: sequence, size n_dim
        The adjusted shape is a multiple of base_shape
    excluded_list: 2D sequence
        Coordinates of exluded pixels of shape (n_dim, n_pixels)
    '''
    for dim in range(0, len(adjusted_shape_inout)):
        # Nothing to adjust, could trip downstream logic
        if sig_shape[dim] <= 1:
            continue
        unique = np.unique(excluded_list[dim])
        # Very many pixels in the way, low chances of a solution
        if len(unique) > sig_shape[dim] / 3:
            adjusted_shape_inout[dim] = sig_shape[dim]
        else:
            stop = sig_shape[dim]
            # Left and right side of an invalid pixel are forbidden
            forbidden = np.concatenate((unique, unique + 1))
            forbidden = forbidden[forbidden <= stop]
            # Invalid pixel at zero is handled separately
            nonzero_filter = forbidden != 0
            m = min(
                stop,
                disjunct_multiplier(
                    excluded=forbidden[nonzero_filter],
                    sig_shape=sig_shape[dim],
                    base_shape=base_shape[dim],
                    target=adjusted_shape_inout[dim]
                )
            )
            # handle zero
            if not np.all(nonzero_filter):
                min_size = max(m, 2)
            else:
                min_size = m
            # Current shape is not a clean multiple
            # of the ideal shape or too small: adjust
            if adjusted_shape_inout[dim] < min_size or adjusted_shape_inout[dim] % m != 0:
                adjusted_shape_inout[dim] = m



from collections import OrderedDict
import pytest

from sparseconverter import (
    BACKENDS, CPU_BACKENDS, CUPY, CUPY_SCIPY_COO, NUMPY, SCIPY_COO, SCIPY_CSC, SCIPY_CSR,
    SPARSE_COO, SPARSE_GCXS, get_backend
)
from libertem.io.dataset.memory import MemoryDataSet
from libertem.udf.base import _execution_plan
from libertem.udf.base import UDF, _get_canonical_backends

from utils import set_device_class


class BackendUDF(UDF):
    def __init__(
        self, preferred_input_dtype=UDF.USE_NATIVE_DTYPE,
        array_backends=UDF.BACKEND_ALL
    ) -> None:
        super().__init__(preferred_input_dtype=preferred_input_dtype, array_backends=array_backends)

    def process_tile(self, tile):
        assert get_backend(tile) in self.get_backends()
        assert get_backend(tile) == self.meta.array_backend
        self.results.backends[0].append((get_backend(tile), self.meta.slice))

    def preprocess(self):
        self.results.backends[0] = []

    def merge(self, dest, src):
        dest.backends[0].append(src.backends[0])

    def get_result_buffers(self):
        return {
            'backends': self.buffer(kind='single', dtype=object, extra_shape=(1,))
        }

    def get_preferred_input_dtype(self):
        '''
        Return the value passed in the constructor.
        '''
        return self.params.preferred_input_dtype

    def get_backends(self):
        return self.params.array_backends


def canonical_plan(execution_plan):
    res = OrderedDict()
    for key, value in execution_plan.items():
        res[key] = frozenset(value)
    return res


def validate(ctx, ds, udfs, available_backends, source_backend, execution_plan, do_run=True):
    all_udfs = set()
    for b, selected_udfs in execution_plan.items():
        for udf in selected_udfs:
            assert b in _get_canonical_backends(udf.get_backends())
            assert b in available_backends
            assert udf not in all_udfs
            all_udfs.add(udf)
    assert set(udfs) == all_udfs
    assert source_backend in ds.array_backends
    if do_run:
        execute(ctx, ds, udfs, available_backends, source_backend, execution_plan)


def execute(ctx, ds, udfs, available_backends, source_backend, execution_plan):
    if source_backend in CPU_BACKENDS and all(key in CPU_BACKENDS for key in execution_plan):
        device_class = 'cpu'
    else:
        device_class = 'cupy'
    class_for_udf = {}
    for key, udfs in execution_plan.items():
        for udf in udfs:
            class_for_udf[udf] = key

    if available_backends != BACKENDS:
        run_backends = tuple(available_backends)
    else:
        run_backends = None

    with set_device_class(device_class):
        res = ctx.run_udf(dataset=ds, udf=udfs, backends=run_backends)
        for r, udf in zip(res, udfs):
            backends = r['backends'].data[0]
            for p in backends:
                for b in p:
                    assert b[0] == class_for_udf[udf]


@pytest.mark.parametrize(
    # Only test the "proper" backends, leave out
    # DOK, NumPy matrix and CUDA
    'left', UDF.BACKEND_ALL
)
@pytest.mark.parametrize(
    # Only test the "proper" backends, leave out
    # DOK, NumPy matrix and CUDA
    'right', UDF.BACKEND_ALL
)
@pytest.mark.parametrize(
    'device_class', (None, 'cpu', 'cuda')
)
def test_simple(lt_ctx, left, right, device_class):
    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=(left, ))
    udf1 = BackendUDF(array_backends=(right,))
    source_backend, execution_plan = _execution_plan(
        udfs=[udf1],
        ds=ds,
        device_class=device_class,
        available_backends=BACKENDS,
    )
    assert source_backend == left
    assert execution_plan == {right: [udf1]}
    validate(
        ctx=lt_ctx,
        ds=ds,
        udfs=[udf1],
        available_backends=BACKENDS,
        source_backend=source_backend,
        execution_plan=execution_plan
    )


@pytest.mark.parametrize(
    'right', UDF.BACKEND_ALL
)
def test_direct_path_ds(lt_ctx, right):

    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=UDF.BACKEND_ALL)
    udf1 = BackendUDF(array_backends=(right,))
    udf2 = BackendUDF(array_backends=UDF.BACKEND_ALL)
    source_backend, execution_plan = _execution_plan(
        udfs=[udf2, udf1],
        ds=ds,
        device_class=None,
        available_backends=BACKENDS,
    )
    assert source_backend == right
    assert canonical_plan(execution_plan) == {right: frozenset([udf1, udf2])}
    validate(
        ctx=lt_ctx,
        ds=ds,
        udfs=[udf1, udf2],
        available_backends=BACKENDS,
        source_backend=source_backend,
        execution_plan=execution_plan
    )


@pytest.mark.parametrize(
    'left', UDF.BACKEND_ALL
)
def test_direct_path_udf(lt_ctx, left):
    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=(left, ))
    udf = BackendUDF(array_backends=UDF.BACKEND_ALL)
    source_backend, execution_plan = _execution_plan(
        udfs=[udf],
        ds=ds,
        device_class=None,
        available_backends=BACKENDS,
    )
    assert source_backend == left
    assert execution_plan == {left: [udf]}
    validate(
        ctx=lt_ctx,
        ds=ds,
        udfs=[udf],
        available_backends=BACKENDS,
        source_backend=source_backend,
        execution_plan=execution_plan
    )


@pytest.mark.parametrize(
    'available', UDF.BACKEND_ALL
)
@pytest.mark.parametrize(
    'device_class', (None, 'cpu', 'cuda')
)
def test_available(lt_ctx, available, device_class):
    # Verify that the choice of backends can be restricted to available options
    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=UDF.BACKEND_ALL)
    udf = BackendUDF(array_backends=UDF.BACKEND_ALL)
    source_backend, execution_plan = _execution_plan(
        udfs=[udf],
        ds=ds,
        device_class=device_class,
        available_backends=(available, ),
    )
    assert source_backend == available
    assert execution_plan == {available: [udf]}
    validate(
        ctx=lt_ctx,
        ds=ds,
        udfs=[udf],
        available_backends=(available, ),
        source_backend=source_backend,
        execution_plan=execution_plan
    )


def test_ds_preference(lt_ctx):
    # Two direct paths and not clear which is better,
    # we pick the first one of the dataset
    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=(SCIPY_COO, NUMPY))
    udf = BackendUDF(array_backends=UDF.BACKEND_ALL)
    source_backend, execution_plan = _execution_plan(
        udfs=[udf],
        ds=ds,
        device_class=None,
        available_backends=BACKENDS,
    )
    assert source_backend == SCIPY_COO
    assert execution_plan == {SCIPY_COO: [udf]}
    validate(
        ctx=lt_ctx,
        ds=ds,
        udfs=[udf],
        available_backends=BACKENDS,
        source_backend=source_backend,
        execution_plan=execution_plan
    )


def test_udf_preference_sparse(lt_ctx):
    # The dataset has both dense and sparse for CPU
    # We include a format with fast conversion from sparse
    # CPU to sparse GPU so that it wins over NumPy
    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=(SPARSE_COO, SCIPY_COO, NUMPY))
    # The UDF has both dense and sparse on GPU and CPU
    udf = BackendUDF(array_backends=(NUMPY, SCIPY_COO, CUPY, CUPY_SCIPY_COO))
    source_backend, execution_plan = _execution_plan(
        udfs=[udf],
        ds=ds,
        # We assign preferentially for GPU to trigger
        # conversion and not pick a direct option
        # unde rthe assumption that the transfer and conversion
        # cost is offset by processing speed
        # If the sparse->sparse converter is inefficient this will
        # actually do dense-> sparse!
        device_class='cuda',
        available_backends=BACKENDS,
    )
    # We use the sparse option
    # SPARSE_COO -> CUPY_SCIPY_COO is currently
    # expensive, so NUMPY would be chosen here
    assert source_backend == SCIPY_COO
    assert execution_plan == {CUPY_SCIPY_COO: [udf]}
    validate(
        ctx=lt_ctx,
        ds=ds,
        udfs=[udf],
        available_backends=BACKENDS,
        source_backend=source_backend,
        execution_plan=execution_plan
    )


def test_udf_preference_dense(lt_ctx):
    # The dataset has only dense for CPU.
    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=(NUMPY, ))
    # The UDF has both dense and sparse on GPU
    # Indicating preference for sparse
    udf = BackendUDF(array_backends=(NUMPY, CUPY_SCIPY_COO, CUPY))
    source_backend, execution_plan = _execution_plan(
        udfs=[udf],
        ds=ds,
        # We assign preferentially for GPU to trigger
        # transfer and not pick the direct NUMPY option
        device_class='cuda',
        available_backends=BACKENDS,
    )
    # We use the dense option to avoid dense->sparse conversion
    assert source_backend == NUMPY
    assert execution_plan == {CUPY: [udf]}
    validate(
        ctx=lt_ctx,
        ds=ds,
        udfs=[udf],
        available_backends=BACKENDS,
        source_backend=source_backend,
        execution_plan=execution_plan
    )


def test_udf_preference_number_order(lt_ctx):
    # The dataset has only dense for CPU.
    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=(NUMPY, ))
    # The UDF has only sparse on CPU
    # Choose by order or occurrence
    udf1 = BackendUDF(array_backends=(SCIPY_CSR, SCIPY_CSC, SCIPY_COO))
    udf2 = BackendUDF(array_backends=(SCIPY_CSR, SCIPY_COO))
    udf3 = BackendUDF(array_backends=(SPARSE_GCXS, SPARSE_COO))
    source_backend, execution_plan = _execution_plan(
        udfs=[udf1, udf2, udf3],
        ds=ds,
        device_class=None,
        available_backends=BACKENDS,
    )
    # We use SCIPY_CSR (fast conversion, more UDFs, before SCIPY_COO),
    # then SPARSE_GCXS (slower conversion), and not SPARSE_COO (less preferred)
    assert source_backend == NUMPY
    assert canonical_plan(execution_plan) == {
        SCIPY_CSR: frozenset([udf1, udf2]),
        SPARSE_GCXS: frozenset([udf3]),
    }
    validate(
        ctx=lt_ctx,
        ds=ds,
        udfs=[udf1, udf2, udf3],
        available_backends=BACKENDS,
        source_backend=source_backend,
        execution_plan=execution_plan
    )


@pytest.mark.parametrize(
    'device_class', (None, 'cpu', 'cuda')
)
@pytest.mark.parametrize(
    'ds_backends', [(CUPY, NUMPY), (NUMPY, CUPY), (NUMPY, ), (CUPY, )]
)
@pytest.mark.parametrize(
    'udf4_backends', [(SCIPY_CSC, SPARSE_GCXS), (SPARSE_GCXS, SCIPY_CSC)]
)
def test_udf_preference_number(lt_ctx, device_class, ds_backends, udf4_backends):
    # The dataset has only dense for CPU or GPU.
    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=ds_backends)
    # The UDFs have only sparse on CPU
    # Choose by number of UDFs and order of occurrence
    udf1 = BackendUDF(array_backends=(SCIPY_CSR, SCIPY_CSC, SCIPY_COO))
    udf2 = BackendUDF(array_backends=(SCIPY_CSR, SPARSE_COO, SCIPY_COO))
    udf3 = BackendUDF(array_backends=(SPARSE_GCXS, SCIPY_COO))
    udf4 = BackendUDF(array_backends=udf4_backends)
    source_backend, execution_plan = _execution_plan(
        udfs=[udf1, udf2, udf3, udf4],
        ds=ds,
        device_class=device_class,
        available_backends=BACKENDS,
    )
    print(source_backend, execution_plan)
    # Prefer options without CPU-GPU conversion
    assert source_backend == NUMPY if NUMPY in ds_backends else CUPY
    # We use SCIPY_COO (more UDFs) despite being less preferred
    # by the UDFs.
    # SCIPY_COO comes first since the converter is faster
    # than SPARSE_GCXS
    # FIXME There might
    # be faster options that require only two conversions as well
    assert canonical_plan(execution_plan) == {
        SCIPY_COO: frozenset([udf1, udf2, udf3]),
        udf4_backends[0]: frozenset([udf4]),
    }
    validate(
        ctx=lt_ctx,
        ds=ds,
        udfs=[udf1, udf2, udf3, udf4],
        available_backends=BACKENDS,
        source_backend=source_backend,
        execution_plan=execution_plan
    )


def test_no_solution_udf():
    ds = MemoryDataSet(datashape=(2, 2, 2, 2), array_backends=UDF.BACKEND_ALL)
    # only GPU for UDF
    udf = BackendUDF(array_backends=(CUPY, ))
    with pytest.raises(RuntimeError):
        source_backend, execution_plan = _execution_plan(
            udfs=[udf],
            ds=ds,
            device_class=None,
            # restrict to CPU only
            available_backends=CPU_BACKENDS,
        )



import numpy as np

from libertem.udf import UDF


class SumUDF(UDF):
    """
    Sum up frames, preserving the signal dimension

    Parameters
    ----------
    dtype : numpy.dtype, optional
        Preferred dtype for computation, default 'float32'. The actual dtype will be determined
        from this value and the dataset's dtype using :meth:`numpy.result_type`.
        See also :ref:`udf dtype`.

    Examples
    --------
    >>> udf = SumUDF()
    >>> result = ctx.run_udf(dataset=dataset, udf=udf)
    >>> np.array(result["intensity"]).shape
    (32, 32)
    """
    def __init__(self, dtype='float32'):
        super().__init__(dtype=dtype)

    def get_preferred_input_dtype(self):
        ''
        return self.params.dtype

    def get_backends(self):
        ''
        return self.BACKEND_ALL

    def get_result_buffers(self):
        ''
        return {
            'intensity': self.buffer(
                kind='sig', dtype=self.meta.input_dtype, where='device'
            )
        }

    def process_tile(self, tile):
        ''
        self.results.intensity[:] += self.forbuf(
            np.sum(tile, axis=0),
            self.results.intensity
        )

    def merge(self, dest, src):
        ''
        dest.intensity[:] += src.intensity

    def merge_all(self, ordered_results):
        ''
        intensity_chunks = [b.intensity for b in ordered_results.values()]
        intensity_sum = np.stack(intensity_chunks, axis=0).sum(axis=0)
        return {'intensity': intensity_sum}



# Based on the SEQ reader of the PIMS project, with the following license:
#
#    Copyright (c) 2013-2014 PIMS contributors
#    https://github.com/soft-matter/pims
#    All rights reserved
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the soft-matter organization nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import struct
import warnings
from typing import Optional

import defusedxml.ElementTree as ET
import numpy as np
import sparse
from ncempy.io.mrc import mrcReader

from libertem.common.math import prod, make_2D_square
from libertem.common import Shape
from libertem.common.messageconverter import MessageConverter
from libertem.io.dataset.base import (
    FileSet, DataSet, BasePartition, DataSetMeta, DataSetException,
    File, IOBackend,
)
from libertem.io.corrections import CorrectionSet

DWORD = 'L'
LONG = 'l'
DOUBLE = 'd'
USHORT = 'H'

HEADER_FIELDS = [
    ('magic', DWORD),
    ('name', '24s'),
    ('version', LONG),
    ('header_size', LONG),
    ('description', '512s'),
    ('width', DWORD),
    ('height', DWORD),
    ('bit_depth', DWORD),
    ('bit_depth_real', DWORD),
    ('image_size_bytes', DWORD),
    ('image_format', DWORD),
    ('allocated_frames', DWORD),
    ('origin', DWORD),
    ('true_image_size', DWORD),
    ('suggested_frame_rate', DOUBLE),
    ('description_format', LONG),
    ('reference_frame', DWORD),
    ('fixed_size', DWORD),
    ('flags', DWORD),
    ('bayer_pattern', LONG),
    ('time_offset_us', LONG),
    ('extended_header_size', LONG),
    ('compression_format', DWORD),
    ('reference_time_s', LONG),
    ('reference_time_ms', USHORT),
    ('reference_time_us', USHORT)
    # More header values not implemented
]

HEADER_SIZE = sum(
    struct.Struct('<' + field[1]).size
    for field in HEADER_FIELDS
)


def _read_header(path, fields):
    with open(path, "rb") as fh:
        fh.seek(0)
        return _unpack_header(fh.read(HEADER_SIZE), fields)


def _unpack_header(header_bytes, fields):
    str_fields = {"name", "description"}
    tmp = dict()
    pos = 0
    for name, fmt in fields:
        header_part = header_bytes[pos:]
        val, size = _unpack_header_field(header_part, fmt)
        if name in str_fields:
            val = _decode_str(val)
        tmp[name] = val
        pos += size

    return tmp


def _unpack_header_field(header_part, fs):
    s = struct.Struct('<' + fs)
    vals = s.unpack(header_part[:s.size])
    if len(vals) == 1:
        return vals[0], s.size
    else:
        return vals, s.size


def _decode_str(str_in):
    end_pos = len(str_in)
    end_mark = b'\x00\x00'
    if end_mark in str_in:
        end_pos = str_in.index(end_mark) + 1
    return str_in[:end_pos].decode("utf16")


def _get_image_offset(header):
    if header['version'] >= 5:
        return 8192
    else:
        return 1024


def xml_map_sizes(bad_pixel_maps):
    """
    returns the sizes and binning values of bad_pixel_maps

    Parameters
    ----------
    bad_pixel_maps: node
        all of the bad pixel map nodes from the root
    """
    map_sizes = []

    for size_map in bad_pixel_maps:
        if "Binning" in size_map.attrib:
            map_sizes.append((int(size_map.attrib['Columns']), int(size_map.attrib['Rows']),
                              int(size_map.attrib["Binning"])))
        else:
            map_sizes.append((int(size_map.attrib['Columns']), int(size_map.attrib['Rows']),
                              int(size_map.attrib.get("Binning", 1))))
    map_rearrange = zip(*map_sizes)
    xy_map_sizes = list(map_rearrange)
    return xy_map_sizes, map_sizes


def xml_unbinned_map_maker(xy_map_sizes):
    """
    returns two list of unbinned sizes(rows and cols) if the size was binned
    return 0 in the list in its place
    Parameters
    ----------
    xy_map_sizes: list of tuple
        bad pixel maps sizes separated by row and col
    """
    unbinned_x = []
    unbinned_y = []
    for map_ind in range(0, len(xy_map_sizes[0])):
        if xy_map_sizes[2][map_ind] < 2:
            unbinned_y.append(xy_map_sizes[0][map_ind])
            unbinned_x.append(xy_map_sizes[1][map_ind])
        else:
            unbinned_y.append(0)
            unbinned_x.append(0)
    return unbinned_x, unbinned_y


def xml_binned_map_maker(xy_map_sizes):
    """
    returns two list of binned sizes(rows and cols) if the size was not binned
    return 0 in the list in its place
    Parameters
    ----------
    xy_map_sizes: list of tuple
        bad pixel maps sizes separated by row and col
    """
    unbinned_x = []
    unbinned_y = []
    for map_ind in range(0, len(xy_map_sizes[0])):
        if xy_map_sizes[2][map_ind] > 1:
            unbinned_y.append(xy_map_sizes[0][map_ind])
            unbinned_x.append(xy_map_sizes[1][map_ind])
        else:
            unbinned_y.append(0)
            unbinned_x.append(0)
    return unbinned_x, unbinned_y


def xml_map_index_selector(used_y):
    """
    returns the bad pixel maps  index with the largest column count

    Parameters
    ----------
    used_y: list
        list of either binned or unbinned y coordinates
    """
    max_y = max(used_y)

    map_index = used_y.index(max_y)
    return map_index


def xml_defect_coord_extractor(bad_pixel_map, map_index, map_sizes):
    """
    returns the chosen bad pixel map's defects

    Parameters
    ----------
    bad_pixel_map: node
        the xml file's root node
    map_index: int
        the index of the correct bad pixel map
    map_sizes: list of tuples
    """
    excluded_rows = []
    excluded_cols = []
    excluded_pixels = []
    for defect in bad_pixel_map.findall('Defect'):
        if len(defect.attrib) == 1:
            defect_attrib_key = defect.attrib.keys()
            if "Rows" in defect_attrib_key:
                split = defect.attrib["Rows"].split('-')
                excluded_rows.append(split)
            if "Row" in defect_attrib_key:
                excluded_rows.append([defect.attrib["Row"]])

            if "Columns" in defect_attrib_key:
                split = defect.attrib["Columns"].split('-')
                excluded_cols.append(split)
            if "Column" in defect_attrib_key:
                excluded_cols.append([defect.attrib["Column"]])
        else:
            excluded_pixels.append([defect.attrib["Column"], defect.attrib["Row"]])

    return {"rows": excluded_rows,
            "cols": excluded_cols,
            "pixels": excluded_pixels,
            "size": (map_sizes[map_index][0], map_sizes[map_index][1])
            }


def xml_defect_data_extractor(root, metadata):
    """
    Parameters
    ----------
    root: node
        the xml file's root node
    metadata: dictionary
        the content of the metadata file as a dictionary
    """
    bad_pixel_maps = root.findall('.//BadPixelMap')
    xy_size, map_sizes = xml_map_sizes(bad_pixel_maps)
    if (metadata['HardwareBinning']) < 2:
        used_x, used_y = xml_unbinned_map_maker(xy_map_sizes=xy_size)
    else:
        used_x, used_y = xml_binned_map_maker(xy_map_sizes=xy_size)
    map_index = xml_map_index_selector(used_y)
    defect_dict = xml_defect_coord_extractor(bad_pixel_maps[map_index], map_index, map_sizes)
    return defect_dict


def bin_array2d(a, binning):
    """
    Parameters
    ----------
    a: np array
    binning: int
        the times we want to bin the array
    """
    sx, sy = a.shape
    sxc = sx // binning * binning
    syc = sy // binning * binning
    # crop:
    ac = a[:sxc, :syc]
    return ac.reshape(ac.shape[0] // binning, binning,
                      ac.shape[1] // binning, binning).sum(3).sum(1)


def array_cropping(arr, start_size, req_size, offsets):
    """
    returns a crop from the original array
    Parameters
    ----------
    arr: np array
        the array we will make the changes on
    start_size: tuple
        the size of the original array
    req_size: tuple
        the size we want to crop to
    offsets: tuple
        the top left coord of the crop
    """
    ac = arr
    if offsets[0] + req_size[0] <= start_size[0] and offsets[1] + req_size[1] <= start_size[1]:
        req_y = int(req_size[0]) // 2
        req_x = int(req_size[1]) // 2
        a = int(offsets[0]) + req_y
        b = int(offsets[1]) + req_x
        ac = ac[(a - req_y):(a + req_y), (b - req_x):(b + req_x)]
    return ac


def xml_generate_map_size(exc_rows, exc_cols, exc_pix, size, metadata):
    """
    This function will be responsible for generating new size based on the parameters.
    returns an np array with the excluded pixels as True

    Parameters
    ----------
    exc_rows: list of lists
        list with the excluded rows as lists which length is 1 if its a single row
        and two if its an interval of rows
    exc_cols: list of lists
    exc_pix: list of lists
    size: tuple
        selected bad pixel maps size
    metadata: dictionary
        the metadata where we specify the image's parameters
    """
    required_size = (metadata["UnbinnedFrameSizeY"], metadata["UnbinnedFrameSizeX"])
    offsets = (metadata["OffsetY"], metadata["OffsetX"])
    bin_value = metadata["HardwareBinning"]
    if bin_value > 1:
        required_size = (required_size[0]//2, required_size[1]//2)
        offsets = (offsets[0]//2, offsets[1]//2)
    dummy_m = np.zeros(size, dtype=bool)
    for row in exc_rows:

        if len(row) == 1:
            dummy_m[int(row[0])] = 1
        else:
            dummy_m[int(row[0]):int(row[1]) + 1] = 1

    for col in exc_cols:

        if len(col) == 1:
            dummy_m[:, int(col[0])] = 1
        else:
            dummy_m[:, int(col[0]):int(col[1]) + 1] = 1

    for pix in exc_pix:
        dummy_m[int(pix[1]), int(pix[0])] = 1

    cropped = array_cropping(dummy_m, start_size=size, req_size=required_size, offsets=offsets)
    return np.array(cropped, dtype=bool)


def xml_processing(tree, metadata_dict):
    data_dict = xml_defect_data_extractor(tree, metadata_dict)
    coord = xml_generate_map_size(data_dict["rows"], data_dict["cols"], data_dict["pixels"],
                                  data_dict["size"], metadata_dict)
    return sparse.COO(coord)


def _load_xml_from_string(xml, metadata):
    tree = ET.fromstring(xml)
    return xml_processing(tree, metadata)


class SEQDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/SEQDatasetParams.schema.json",
        "title": "SEQDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "SEQ"},
            "path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
            "io_backend": {
                "enum": IOBackend.get_supported(),
            }
        },
        "required": ["type", "path", "nav_shape"],
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path", "nav_shape"]
        }
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


class SEQFileSet(FileSet):
    pass


class SEQDataSet(DataSet):
    """
    Read data from Norpix SEQ files.

    Examples
    --------

    >>> ds = ctx.load("seq", path="/path/to/file.seq", nav_shape=(1024, 1024))  # doctest: +SKIP

    Parameters
    ----------

    path
        Path to the .seq file
    nav_shape: tuple of int
        A n-tuple that specifies the size of the navigation region ((y, x), but
        can also be of length 1 for example for a line scan, or length 3 for
        a data cube, for example)
    sig_shape: tuple of int, optional
        Signal/detector size (height, width)
    sync_offset: int, optional
        If positive, number of frames to skip from start.
        If negative, number of blank frames to insert at start.

    Note
    ----

    Dark and gain reference are loaded from MRC files with the same root as the
    SEQ file and the extensions :code:`.dark.mrc` and :code:`.gain.mrc`, i.e.
    :code:`/path/to/file.dark.mrc` and :code:`/path/to/file.gain.mrc` if they are present.

    .. versionadded:: 0.8.0

    Dead pixels are read from an XML file with the same root as the
    SEQ file and the extension :code:`.Config.Metadata.xml`, i.e.
    :code:`/path/to/file.Config.Metadata.xml` in the above example if both this file and
    :code:`/path/to/file.metadata` are present.

    See :ref:`corrections` for more information on how to change or disable corrections.

    FIXME find public documentation of the XML format and dark/gain maps.
    """

    def __init__(
        self,
        path: str,
        scan_size: Optional[tuple[int, ...]] = None,
        nav_shape: Optional[tuple[int, ...]] = None,
        sig_shape: Optional[tuple[int, ...]] = None,
        sync_offset: int = 0,
        io_backend=None,
    ):
        super().__init__(io_backend=io_backend)
        self._path = path
        # There might be '.seq.seq' and '.seq' in the wild
        # See https://github.com/LiberTEM/LiberTEM/issues/1120
        # We first try if '.seq.seq' matches, then '.seq'
        name, ext = os.path.splitext(path)
        name2, ext2 = os.path.splitext(name)
        if ext.lower() == '.seq' and ext2.lower() == '.seq':
            self._basename = name2
        elif ext.lower() == '.seq':
            self._basename = name
        else:
            self._basename = path
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset
        # handle backwards-compatability:
        if scan_size is not None:
            warnings.warn(
                "scan_size argument is deprecated. please specify nav_shape instead",
                FutureWarning
            )
            if nav_shape is not None:
                raise ValueError("cannot specify both scan_size and nav_shape")
            self._nav_shape = tuple(scan_size)
        if self._nav_shape is None:
            raise TypeError("missing 1 required argument: 'nav_shape'")
        self._header = None
        self._meta = None
        self._footer_size = None
        self._filesize = None
        self._dark = None
        self._gain = None
        self._excluded_pixels = None

    def _do_initialize(self):
        header = self._header = _read_header(self._path, HEADER_FIELDS)
        self._image_offset = _get_image_offset(header)
        if header['version'] >= 5:  # StreamPix version 6
            # Timestamp = 4-byte unsigned long + 2-byte unsigned short (ms)
            #   + 2-byte unsigned short (us)
            self._timestamp_micro = True
        else:  # Older versions
            self._timestamp_micro = False
        try:
            dtype = np.dtype('uint%i' % header['bit_depth'])
        except TypeError:
            raise DataSetException("unsupported bit depth: %s" % header['bit_depth'])
        frame_size_bytes = header['width'] * header['height'] * dtype.itemsize
        self._footer_size = header['true_image_size'] - frame_size_bytes
        self._filesize = os.stat(self._path).st_size
        self._image_count = int((self._filesize - self._image_offset) / header['true_image_size'])

        if self._sig_shape is None:
            self._sig_shape = tuple((header['height'], header['width']))
        elif int(prod(self._sig_shape)) != (header['height'] * header['width']):
            raise DataSetException(
                "sig_shape must be of size: %s" % (header['height'] * header['width'])
            )

        self._nav_shape_product = int(prod(self._nav_shape))
        self._sync_offset_info = self.get_sync_offset_info()
        shape = Shape(self._nav_shape + self._sig_shape, sig_dims=len(self._sig_shape))

        self._meta = DataSetMeta(
            shape=shape,
            raw_dtype=dtype,
            dtype=dtype,
            metadata=header,
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        self._maybe_load_dark_gain()
        return self

    def _maybe_load_mrc(self, path):
        if not os.path.exists(path):
            return None
        data_dict = mrcReader(path)
        return np.squeeze(data_dict['data'])

    def _load_xml_from_file(self):
        xml_path = self._basename + ".seq.Config.Metadata.xml"
        meta_path = self._basename + ".seq.metadata"
        if not os.path.exists(xml_path):
            return None
        if not os.path.exists(meta_path):
            return None
        else:
            tree = ET.parse(xml_path)
            root = tree.getroot()
            with open(meta_path, mode="rb") as file:
                met = file.read()
            metdata_keys = ['DEMetadataSize', 'DEMetadataVersion', 'UnbinnedFrameSizeX',
                            'UnbinnedFrameSizeY', 'OffsetX', 'OffsetY', 'HardwareBinning',
                            'Bitmode', 'FrameRate', 'RotationMode',
                            'FlipMode', 'OkraMode']
            metadata = dict(zip(metdata_keys, struct.unpack_from('iiiiiiiiiii?', met, 282)))
            return xml_processing(root, metadata)

    def _maybe_load_dark_gain(self):
        self._dark = self._maybe_load_mrc(self._basename + ".seq.dark.mrc")
        self._gain = self._maybe_load_mrc(self._basename + ".seq.gain.mrc")
        self._excluded_pixels = self._load_xml_from_file()

    def get_correction_data(self):
        return CorrectionSet(
            dark=self._dark,
            gain=self._gain,
            excluded_pixels=self._excluded_pixels,
        )

    def initialize(self, executor):
        return executor.run_function(self._do_initialize)

    def get_diagnostics(self):
        header = self._header
        return [
                   {"name": k, "value": str(v)}
                   for k, v in header.items()
               ] + [
                   {"name": "Footer size",
                    "value": str(self._footer_size)},
                   {"name": "Dark frame included",
                    "value": str(self._dark is not None)},
                   {"name": "Gain map included",
                    "value": str(self._gain is not None)},
               ]

    @property
    def meta(self):
        return self._meta

    @classmethod
    def get_msg_converter(cls):
        return SEQDatasetParams

    @classmethod
    def _do_detect_params(cls, path):
        try:
            header = _read_header(path, HEADER_FIELDS)
            if header['magic'] != 0xFEED:
                return False
            image_offset = _get_image_offset(header)
            filesize = os.stat(path).st_size
            image_count = int(
                (filesize - image_offset) / header['true_image_size']
            )
            sig_shape = tuple((header['height'], header['width']))
            return {
                "parameters": {
                    "path": path,
                    "nav_shape": make_2D_square((image_count,)),
                    "sig_shape": sig_shape,
                },
                "info": {
                    "image_count": image_count,
                    "native_sig_shape": sig_shape,
                }
            }
        except (OSError, UnicodeDecodeError):
            return False

    @classmethod
    def detect_params(cls, path, executor):
        try:
            return executor.run_function(cls._do_detect_params, path)
        except Exception as e:
            raise DataSetException(repr(e)) from e

    @classmethod
    def get_supported_extensions(cls):
        return {"seq"}

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    def _get_fileset(self):
        return SEQFileSet(files=[
            File(
                path=self._path,
                start_idx=0,
                end_idx=self._image_count,
                native_dtype=self._meta.raw_dtype,
                sig_shape=self._meta.shape.sig,
                frame_footer=self._footer_size,
                frame_header=0,
                file_header=self._image_offset,
            )
        ], frame_footer_bytes=self._footer_size)

    def check_valid(self):
        if self._header['magic'] != 0xFEED:
            raise DataSetException('The format of this .seq file is unrecognized')
        if self._header['compression_format'] != 0:
            raise DataSetException('Only uncompressed images are supported in .seq files')
        if self._header['image_format'] != 100:
            raise DataSetException('Non-monochrome images are not supported')

    def get_cache_key(self):
        return {
            "path": self._path,
            "shape": self.shape,
            "sync_offset": self._sync_offset,
        }

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield BasePartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )

    def __repr__(self):
        return f"<SEQDataSet of {self.dtype} shape={self.shape}>"



import pytest
import numpy as np

from libertem.udf.base import UDFRunner, UDF
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


class PixelsumUDF(UDF):
    def get_result_buffers(self):
        return {
            'pixelsum': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_frame(self, frame):
        assert frame.shape == (16, 16)
        assert self.results.pixelsum.shape == (1,)
        self.results.pixelsum[:] = np.sum(frame)


@pytest.mark.asyncio
async def test_async_run_for_dset(async_executor):
    data = _mk_random(size=(16 * 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    pixelsum = PixelsumUDF()
    roi = np.zeros((256,), dtype=bool)
    runner = UDFRunner([pixelsum])

    udf_iter = runner.run_for_dataset_async(
        dataset, async_executor, roi=roi, cancel_id="42"
    )

    async for udf_results in udf_iter:
        udf_results.buffers
        pass
    assert "udf_results" in locals(), "must yield at least one result"



import os
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Optional
import contextlib
import uuid

from opentelemetry import trace, context
from opentelemetry.trace import NonRecordingSpan, SpanContext

if TYPE_CHECKING:
    from libertem.io.dataset.base import Partition

logger = logging.getLogger()


def maybe_setup_tracing(
    service_name: str,
    service_id: Optional[str] = None,
    otlp_endpoint: Optional[str] = None,
):
    """
    Set up tracing if the OTEL_ENABLE environment variable is set.
    Only call this once per process.

    Please see :func:`setup_tracing` for a description of the parameters.
    """
    if os.environ.get('OTEL_ENABLE') is not None:
        return setup_tracing(service_name, service_id, otlp_endpoint)


def setup_tracing(
    service_name: str,
    service_id: Optional[str] = None,
    otlp_endpoint: Optional[str] = None,
):
    """
    Set up tracing and span exporting. Call once per process.

    Parameters
    ----------

    service_name
        The service name which will appear in the trace

    service_id
        The service id, if there are multiple instances of the same service
        (i.e. worker processes). Should stay the same over the runtime of
        the program, but can be different when "major lifecycle events" happen,
        for example program restarts

    otlp_endpoint
        Set up exporting of tracing spans via OTLP. If this is not set, this
        will default to the value of the OTEL_EXPORTER_OTLP_ENDPOINT
        environment variable, or fall back to http://localhost:4317
    """
    if otlp_endpoint is None:
        otlp_endpoint = os.environ.get("OTEL_EXPORTER_OTLP_ENDPOINT")
    if otlp_endpoint is None:
        otlp_endpoint = "http://localhost:4317"
    try:
        from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
        from opentelemetry.sdk.resources import Resource
        from opentelemetry.sdk.trace import TracerProvider
        from opentelemetry.sdk.trace.export import BatchSpanProcessor
        # from opentelemetry.sdk.trace.export import ConsoleSpanExporter
    except ImportError:
        logger.warn("could not import opentelemetry library, disabling tracing")
        return

    if service_id is None:
        service_id = str(uuid.uuid4())
    provider = TracerProvider(resource=Resource.create({
        "service.name": service_name,
        "service.instance.id": service_id,
    }))
    otlp_exporter = OTLPSpanExporter(endpoint=otlp_endpoint, insecure=True)
    # console_exporter = ConsoleSpanExporter()
    processor = BatchSpanProcessor(otlp_exporter)
    provider.add_span_processor(processor)
    trace.set_tracer_provider(provider)


@contextlib.contextmanager
def attach_to_parent(span_context: SpanContext):
    """
    Attach to an existing span, such that new spans created will be parented
    to it. This is useful for serializing the `SpanContext` to a different
    process, possibly running on a different node, and keeping the tracing
    relation.

    See also: https://opentelemetry.io/docs/instrumentation/python/cookbook/
    """
    trace_context = trace.set_span_in_context(NonRecordingSpan(span_context))
    token = context.attach(trace_context)
    try:
        yield
    finally:
        context.detach(token)


def add_partition_to_span(partition: "Partition") -> None:
    """
    Add Partition metadata (start_idx, end_idx) to the current tracing span
    """
    span = trace.get_current_span()
    slice_ = partition.slice
    span.set_attributes({
        "libertem.partition.start_idx": slice_.origin[0],
        "libertem.partition.end_idx": slice_.origin[0] + slice_.shape[0],
    })


class TracedThreadPoolExecutor(ThreadPoolExecutor):
    """
    A :code:`~concurrent.futures.ThreadPoolExecutor` that propagates the submit
    time tracing context to the worker threads.

    >>> tracer = trace.get_tracer(__name__)
    >>> pool = TracedThreadPoolExecutor(tracer)
    >>> fut = pool.submit(...)
    """
    def __init__(self, tracer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tracer = tracer

    def _in_parent_span(self, span_context: SpanContext, fn, *args, **kwargs):
        with attach_to_parent(span_context):
            return fn(*args, **kwargs)

    def submit(self, __fn, *args, **kwargs):
        if context.get_current():
            span = trace.get_current_span()
            span_context = span.get_span_context()
            return super().submit(
                lambda: self._in_parent_span(span_context, __fn, *args, **kwargs)
            )
        return super().submit(__fn, *args, **kwargs)
